# ----------------------------------------------------------------------
# Copyright (c) 2024 Rafael Gonzalez.
#
# See the LICENSE file for details
# ----------------------------------------------------------------------

# --------------------
# System wide imports
# -------------------

import asyncio
import logging
from datetime import datetime
from argparse import Namespace, ArgumentParser

# -------------------
# Third party imports
# -------------------

from lica.sqlalchemy import sqa_logging
from lica.asyncio.cli import execute

# --------------
# local imports
# -------------

from .. import __version__
from .util import parser as prs
from ..dao import engine
from ..controller.exporter import Controller as Exporter


# ----------------
# Module constants
# ----------------

TSTAMP_FMT = "%Y-%m-%dT%H:%M:%S"

# -----------------------
# Module global variables
# -----------------------

# get the module logger
log = logging.getLogger(__name__.split(".")[-1])


# -----------------
# Auxiliar function
# -----------------


def this_year(value: datetime | None) -> datetime:
    """Date & time validator for the command line interface"""
    return value or datetime.now().replace(
        month=1, day=1, hour=0, minute=0, second=0, microsecond=0
    )


def next_year(value: datetime | None) -> datetime:
    """Date & time validator for the command line interface"""
    return value or datetime.now().replace(
        month=12, day=31, hour=23, minute=59, second=59, microsecond=0
    )


# -----------------
# CLI API functions
# -----------------


async def cli_session_export(args: Namespace) -> None:
    log.info("exporting session %s", args.session)
    log.info("exporting to directory %s", args.base_dir)
    assert isinstance(args.session, datetime)
    exporter = Exporter(
        base_dir=args.base_dir,
        begin_tstamp=args.session,
        end_tstamp=args.session,
        filename_prefix="session",
    )
    N = await exporter.query_nsummaries()
    if N > 0:
        summaries = await exporter.query_summaries()
        await asyncio.to_thread(exporter.export_summaries, summaries)
        rounds = await exporter.query_rounds()
        await asyncio.to_thread(exporter.export_rounds, rounds)
        samples = await exporter.query_samples()
        await asyncio.to_thread(exporter.export_samples, samples)
        zip_file_path = await asyncio.to_thread(exporter.pack)
        log.info("zipped file in  %s", zip_file_path)
    else:
        log.warn("No calibration session found for %s", args.session)
    return


async def cli_session_count(args: Namespace) -> None:
    args.since = args.since or this_year(args.since)
    args.until = args.until or next_year(args.until)
    exporter = Exporter(
        base_dir=args.base_dir,
        begin_tstamp=args.since,
        end_tstamp=args.until,
        filename_prefix="count",
    )
    N = await exporter.query_nsummaries(mode=args.mode)
    log.info("%d calibrations made between %s and %s", N, args.since, args.until)
    if args.detailed:
        # Sort result by calibration date
        summaries = sorted(
            await exporter.query_summaries(args.mode), key=lambda x: x[5]
        )  # Sort by session (calibration date)
        await asyncio.to_thread(exporter.export_summaries, summaries)
    return


def add_args(parser: ArgumentParser):
    subparser = parser.add_subparsers(dest="command", required=True)
    p = subparser.add_parser(
        "single",
        parents=[prs.sess(), prs.bdir()],
        help="Export a single calibration session to CSV files",
    )
    p.set_defaults(func=cli_session_export)
    p = subparser.add_parser(
        "count",
        parents=[prs.trange(), prs.bdir(), prs.mode(), prs.detailed()],
        help="Count number of calibrations from a given time range",
    )
    p.set_defaults(func=cli_session_count)


async def cli_main(args: Namespace) -> None:
    sqa_logging(args)
    await args.func(args)
    await engine.dispose()


def main():
    """The main entry point specified by pyproject.toml"""
    execute(
        main_func=cli_main,
        add_args_func=add_args,
        name=__name__,
        version=__version__,
        description="Additional tools",
    )



# ----------------------------------------------------------------------
# Copyright (c) 2024 Rafael Gonzalez.
#
# See the LICENSE file for details
# ----------------------------------------------------------------------

# --------------------
# System wide imports
# -------------------

import os
import csv
import glob
import zipfile
import logging
import itertools
import contextlib

import ssl
import smtplib

from email import encoders
from email.mime.base import MIMEBase
from email.mime.multipart import MIMEMultipart
from email.mime.text import MIMEText


from datetime import datetime
from typing import Sequence, Tuple, Any

# -------------------
# Third party imports
# -------------------

import aiohttp
from sqlalchemy import select, func, cast, Integer


from zptessdao.asyncio import SummaryView, RoundsView, SampleView, Config, Batch
from zptessdao.constants import Calibration

# --------------
# local imports
# -------------

from ..dao import Session


SUMMARY_EXPORT_HEADERS = (
    "Model",
    "Name",
    "MAC",
    "Firmware",
    "Sensor",
    "Calibration Date (UTC)",
    "Calibration",
    "Cal. SW. Version",
    "Ref. Mag.",
    "Ref. Freq.",
    "Test Mag.",
    "test freq.",
    "Ref-Test Mag. Diff.",
    "Raw ZP",
    "ZP Offset",
    "Final ZP",
    "Prev. ZP",
    "Filter",
    "Plug",
    "Box",
    "Collector",
    "Author",
    "Comment",
)

ROUND_EXPORT_HEADERS = (
    "Model",
    "Name",
    "MAC",
    "Session (UTC)",
    "Role",
    "Round",
    "Freq (Hz)",
    "\u03c3 (Hz)",
    "Mag",
    "ZP",
    "# Samples",
    "\u0394T (s.)",
)

SAMPLE_EXPORT_HEADERS = (
    "Model",
    "Name",
    "MAC",
    "Session (UTC)",
    "Role",
    "Round",
    "Timestamp",
    "Freq (Hz)",
    "Box Temp (\u2103)",
    "Sequence #",
)

# -----------------------
# Module global variables
# -----------------------

# get the module logger
log = logging.getLogger(__name__.split(".")[-1])


# -------------------
# Auxiliary functions
# -------------------


# Adapted From https://realpython.com/python-send-email/
def email_send(
    subject: str,
    body: str,
    sender: str,
    receivers: str,
    attachment: str,
    host: str,
    port: str,
    password: str,
    confidential: bool = False,
):
    msg_receivers = receivers
    receivers = receivers.split(sep=",")
    message = MIMEMultipart()
    message["Subject"] = subject
    # Create a multipart message and set headers
    if confidential:
        message["From"] = sender
        message["To"] = sender
        message["Bcc"] = msg_receivers
    else:
        message["From"] = sender
        message["To"] = msg_receivers

    # Add body to email
    message.attach(MIMEText(body, "plain"))

    # Open file in binary mode
    with open(attachment, "rb") as fd:
        # Add file as application/octet-stream
        # Email client can usually download this automatically as attachment
        part = MIMEBase("application", "octet-stream")
        part.set_payload(fd.read())

    # Encode file in ASCII characters to send by email
    encoders.encode_base64(part)

    # Add header as key/value pair to attachment part
    part.add_header(
        "Content-Disposition",
        f"attachment; filename= {os.path.basename(attachment)}",
    )
    # Add attachment to message and convert message to string
    message.attach(part)
    # Log in to server using secure context and send email
    context = ssl.create_default_context()
    with smtplib.SMTP(host, port) as server:
        server.ehlo()  # Can be omitted
        server.starttls(context=context)
        server.ehlo()  # Can be omitted
        server.login(sender, password)
        server.sendmail(sender, receivers, message.as_string())


# -----------------
# Auxiliary classes
# -----------------


class Controller:
    def __init__(
        self,
        base_dir: str,
        filename_prefix: str,
        begin_tstamp: datetime = None,
        end_tstamp: datetime = None,
    ):
        self.begin_tstamp = begin_tstamp
        self.end_tstamp = end_tstamp
        self.base_dir = base_dir if os.path.isabs(base_dir) else os.path.abspath(base_dir)
        self.filename_prefix = filename_prefix

    # ----------
    # Public API
    # ----------


    async def query_nsummaries(self, mode: Calibration = None) -> int:
        """Return the number of summaries from a time span, even if they are not updated"""
        summaries = await self.query_summaries(mode)
        return len(summaries)

    async def query_summaries(self, mode: Calibration = None) -> Sequence[Tuple[Any]]:
        async with Session() as session:
            async with session.begin():
                t0 = self.begin_tstamp
                t1 = self.end_tstamp
                q = select(
                    SummaryView.model,
                    SummaryView.name,
                    SummaryView.mac,
                    SummaryView.firmware,
                    SummaryView.sensor,
                    SummaryView.session,
                    SummaryView.calibration,
                    SummaryView.calversion,
                    SummaryView.ref_mag,
                    SummaryView.ref_freq,
                    SummaryView.test_freq,
                    SummaryView.test_mag,
                    SummaryView.mag_diff,
                    SummaryView.raw_zero_point,
                    SummaryView.zp_offset,
                    SummaryView.zero_point,
                    SummaryView.prev_zp,
                    SummaryView.filter,
                    SummaryView.plug,
                    SummaryView.box,
                    SummaryView.collector,
                    SummaryView.author,
                    SummaryView.comment,
                )
                if mode is not None:
                    q = q.where(SummaryView.calibration == mode)
                if t0 is not None:
                    q = q.where(
                        SummaryView.session.between(t0, t1),
                        SummaryView.upd_flag == True,  # noqa: E712
                    ).order_by(cast(func.substr(SummaryView.name, 6), Integer), SummaryView.session)
                else:
                    q = q.where(
                        SummaryView.name.like("stars%"),
                        SummaryView.upd_flag == True,  # noqa: E712
                    ).order_by(cast(func.substr(SummaryView.name, 6), Integer), SummaryView.session)
                summaries = (await session.execute(q)).all()
                summaries = self._filter_latest_summary(summaries)
        return summaries

    async def query_rounds(self) -> Sequence[Tuple[Any]]:
        async with Session() as session:
            async with session.begin():
                t0 = self.begin_tstamp
                t1 = self.end_tstamp
                q = (
                    select(
                        RoundsView.model,
                        RoundsView.name,
                        RoundsView.mac,
                        RoundsView.session,
                        RoundsView.role,
                        RoundsView.round,
                        RoundsView.freq,
                        RoundsView.stddev,
                        RoundsView.mag,
                        RoundsView.zero_point,
                        RoundsView.nsamples,
                        RoundsView.duration,
                    )
                    # complicated filter because stars3 always has upd_flag = False
                    .where(
                        RoundsView.session.between(t0, t1)
                        & (
                            (RoundsView.upd_flag == True)  # noqa: E712
                            | ((RoundsView.upd_flag == False) & (RoundsView.name == "stars3"))  # noqa: E712
                        )
                    )
                    .order_by(RoundsView.session, RoundsView.round)
                )
                rounds = (await session.execute(q)).all()
        return rounds

    async def query_samples(self) -> Sequence[Tuple[Any]]:
        async with Session() as session:
            async with session.begin():
                t0 = self.begin_tstamp
                t1 = self.end_tstamp
                q = (
                    select(
                        SampleView.model,
                        SampleView.name,
                        SampleView.mac,
                        SampleView.session,
                        SampleView.role,
                        SampleView.round,
                        SampleView.tstamp,
                        SampleView.freq,
                        SampleView.temp_box,
                        SampleView.seq,
                    )
                    # complicated filter because stars3 always has upd_flag = False
                    .where(
                        SampleView.session.between(t0, t1)
                        & (
                            (SampleView.upd_flag == True)  # noqa: E712
                            | ((SampleView.upd_flag == False) & (SampleView.name == "stars3"))  # noqa: E712
                        )
                    )
                    .order_by(SampleView.session, SampleView.round, SampleView.tstamp)
                )
                rounds = (await session.execute(q)).all()
        return rounds

    def export_summaries(self, summaries: Sequence[Tuple[Any]]) -> None:
        csv_path = os.path.join(self.base_dir, f"summary_{self.filename_prefix}.csv")
        log.info("exporting %s", os.path.basename(csv_path))
        with open(csv_path, "w") as csv_file:
            csv_writer = csv.writer(csv_file, delimiter=";")
            csv_writer.writerow(SUMMARY_EXPORT_HEADERS)
            for summary in summaries:
                csv_writer.writerow(summary)

    def export_rounds(self, rounds: Sequence[Tuple[Any]]) -> None:
        csv_path = os.path.join(self.base_dir, f"rounds_{self.filename_prefix}.csv")
        log.info("exporting %s", os.path.basename(csv_path))
        with open(csv_path, "w") as csv_file:
            csv_writer = csv.writer(csv_file, delimiter=";")
            csv_writer.writerow(ROUND_EXPORT_HEADERS)
            for round_ in rounds:
                csv_writer.writerow(round_)

    def export_samples(self, samples: Sequence[Tuple[Any]]) -> None:
        csv_path = os.path.join(self.base_dir, f"samples_{self.filename_prefix}.csv")
        log.info("exporting %s", os.path.basename(csv_path))
        with open(csv_path, "w") as csv_file:
            csv_writer = csv.writer(csv_file, delimiter=";")
            csv_writer.writerow(SAMPLE_EXPORT_HEADERS)
            for sample in samples:
                csv_writer.writerow(sample)

    def pack(self) -> str:
        """Pack all files in the ZIP file given by options"""
        parent_dir = os.path.dirname(self.base_dir)
        zip_file = os.path.join(parent_dir, self.filename_prefix + ".zip")
        log.info("Creating ZIP File: '%s'", zip_file)
        file_paths = [
            os.path.join(os.path.basename(self.base_dir), fname)
            for fname in glob.glob("*.csv", root_dir=self.base_dir)
        ]
        with contextlib.chdir(parent_dir):
            with zipfile.ZipFile(zip_file, "w") as myzip:
                for myfile in file_paths:
                    myzip.write(myfile)
        return zip_file

    async def check_internet(self) -> bool:
        result = True
        timeout = aiohttp.ClientTimeout(total=5)  # 5 seconds timeout
        async with aiohttp.ClientSession(timeout=timeout) as session:
            try:
                async with session.get("http://www.google.com") as response:
                    status = response.status
                    log.info("Connected to Internet. Status code: %s", status)
            except Exception as e:
                log.exception(e)
                result = False
        return result

    async def load_email_config(self) -> None:
        # Read email configuration
        smtp_keys = set(("host", "port", "sender", "password", "receivers"))
        async with Session() as session:
            async with session.begin():
                q = select(Config).where(Config.section == "smtp").order_by(Config.prop)
                configs = (await session.scalars(q)).all()
        properties = set(cfg.prop for cfg in configs)
        if properties != smtp_keys:
            missing = smtp_keys - properties
            raise Exception("Missing properies in the database: %s", missing)
        self.mail_cfg = dict(map(lambda cfg: (cfg.prop, cfg.value), configs))
        self.mail_cfg["port"] = int(self.mail_cfg["port"])

    def send_email(self, zip_file_path: str) -> bool:
        try:
            email_sent = True
            email_send(
                subject=f"[STARS4ALL] TESS calibration data from {self.begin_tstamp} to {self.end_tstamp}",
                body="Find attached hereafter the summary, rounds and samples from this calibration batch",
                sender=self.mail_cfg["sender"],
                receivers=self.mail_cfg["receivers"],
                attachment=zip_file_path,
                host=self.mail_cfg["host"],
                port=self.mail_cfg["port"],
                password=self.mail_cfg["password"],
            )
        except Exception as e:
            email_sent = False
            log.exception(e)
        return email_sent

    async def update_batch(self, batch: Batch, email_sent: bool) -> None:
        async with Session() as session:
            async with session.begin():
                batch.email_sent = email_sent
                session.add(batch)

    # ---------------
    # Private methods
    # ---------------

    def _filter_latest_summary(self, summaries: Sequence[Tuple[Any]]) -> Sequence[Tuple[Any]]:
        # group by photometer name
        grouped = itertools.groupby(summaries, key=lambda summary: summary[1])
        result = list()
        for name, group in grouped:
            group = tuple(group)
            result.append(
                group[-1]
            )  # Since they are sorted by ascending order, choose the last one
            if len(group) > 1:
                log.warn("%s has %d summaries, choosing the most recent session", name, len(group))
        return result



import statistics


from typing import  Sequence, Tuple

from zptessdao.constants import CentralTendency

def mode(sequence: Sequence) -> float:
    try:
        result = statistics.multimode(sequence)
        if len(result) != 1:     # To make it compatible with my previous software
            raise statistics.StatisticsError
        result = result[0]
    except AttributeError: # Previous to Python 3.8
        result = statistics.mode(sequence)
    return result

def best(sequence: Sequence) -> Tuple[CentralTendency, float]:
    try:
        result = mode(sequence)
        central = CentralTendency.MODE
    except statistics.StatisticsError:
        result = statistics.median_low(sequence)
        central = CentralTendency.MEDIAN
    return central, result


# ----------------------------------------------------------------------
# Copyright (c) 2022
#
# See the LICENSE file for details
# see the AUTHORS file for authors
# ----------------------------------------------------------------------

# ---------------------
# Third party libraries
# ---------------------


SERIAL_PORT_PREFIX = "/dev/ttyUSB"

# TESS-W data

TEST_IP = "192.168.4.1"
TEST_TCP_PORT = 23
TEST_UDP_PORT = 2255
TEST_SERIAL_PORT = 0
TEST_BAUD = 9600

# Timestamp format
TSTAMP_FORMAT = "%Y-%m-%dT%H:%M:%S.%f"

# Condensed timestamp
TSTAMP_SESSION_FMT = "%Y-%m-%dT%H:%M:%S"

# ZP of stars3
ZP_ABS = 20.44


type FreqSequence = list[float, ...]
type TimeSequence = list[float, ...]
type NameSequence = list[str, ...]



# ----------------------------------------------------------------------
# Copyright (c) 2024 Rafael Gonzalez.
#
# See the LICENSE file for details
# ----------------------------------------------------------------------

# --------------------
# System wide imports
# -------------------

import logging
import statistics
import collections
from datetime import datetime
from typing import Tuple, Mapping, Set, Any

# -------------------
# Third party imports
# -------------------

from zptessdao.constants import CentralTendency

# --------------
# local imports
# -------------


# ----------------
# Module constants
# ----------------

Message = Mapping[str, Any]

# -----------------------
# Module global variables
# -----------------------

# get the root logger
log = logging.getLogger(__name__.split(".")[-1])

# -------
# Classes
# -------
# This meant for easy sample de-duplication
# which are usually shared between rounds


class UniqueReading(dict):
    """A hashable, subclaased dictionary based on the "tstamp" keyword and value"""

    def __init__(self, *args, **kwargs):
        self.update(*args, **kwargs)

    def __getitem__(self, key):
        return dict.__getitem__(self, key)

    def __setitem__(self, key, val):
        dict.__setitem__(self, key, val)

    def __repr__(self):
        return "%s(%s)" % (type(self).__name__, dict.__repr__(self))

    def __hash__(self):
        return int(dict.__getitem__(self, "tstamp").timestamp() * 1000)

    def update(self, *args, **kwargs):
        for k, v in dict(*args, **kwargs).items():
            self[k] = v


class RingBuffer:
    def __init__(
        self,
        capacity: int = 75,
        central: CentralTendency = CentralTendency.MEDIAN,
    ):
        self._buffer = collections.deque([], capacity)
        self._central = central
        if central == CentralTendency.MEDIAN:
            self._central_func = statistics.median_low
        elif central == CentralTendency.MEAN:
            self._central_func = statistics.fmean
        elif central == CentralTendency.MODE:
            self._central_func = statistics.mode

    def __len__(self) -> int:
        return len(self._buffer)

    def __getitem__(self, i: int) -> Message:
        return self._buffer[i]

    def capacity(self) -> int:
        return self._buffer.maxlen

    def pop(self) -> Message:
        return self._buffer.popleft()

    def append(self, item: Message) -> None:
        self._buffer.append(item)

    def copy(self) -> Set[Message]:
        return set(UniqueReading(item) for item in self._buffer)

    def intervals(self) -> Tuple[datetime, datetime]:
        return self._buffer[0]["tstamp"], self._buffer[-1]["tstamp"]

    def statistics(self) -> Tuple[float, float]:
        frequencies = tuple(item["freq"] for item in self._buffer)
        central = self._central_func(frequencies)
        stdev = statistics.stdev(frequencies, central)
        return central, stdev



# ----------------------------------------------------------------------
# Copyright (c) 2024 Rafael Gonzalez.
#
# See the LICENSE file for details
# ----------------------------------------------------------------------

# --------------------
# System wide imports
# -------------------

import os
import logging
import zipfile

from datetime import datetime, timezone
from typing import Tuple, Iterable

# ---------------------------
# Third-party library imports
# ----------------------------

from sqlalchemy import select, delete, func

from zptessdao.asyncio import Batch, SummaryView

# --------------
# local imports
# -------------

from ...dao import Session

# -----------------------
# Module global variables
# -----------------------

# get the module logger
log = logging.getLogger(__name__.split(".")[-1])


# ------------------
# Auxiliar functions
# ------------------


def get_paths(directory: str) -> Iterable:
    """Get all file paths in a list"""
    file_paths = list()
    # crawling through directory and subdirectories
    for root, directories, files in os.walk(directory):
        root = os.path.basename(root)  # Needs a change of cwd later on if we do this
        for filename in files:
            filepath = os.path.join(root, filename)
            file_paths.append(filepath)
    return file_paths


def pack(base_dir: str, zip_file: str):
    """Pack all files in the ZIP file given by options"""
    paths = get_paths(base_dir)
    log.info(f"Creating ZIP File: '{os.path.basename(zip_file)}'")
    with zipfile.ZipFile(zip_file, "w") as myzip:
        for myfile in paths:
            myzip.write(myfile)


class Controller:
    def __init__(self):
        pass

    async def open(self, comment: str | None) -> datetime:
        begin_tstamp = datetime.now(timezone.utc).replace(microsecond=0)
        async with Session() as session:
            async with session.begin():
                await self._assert_closed(session)
                batch = Batch(
                    begin_tstamp=begin_tstamp, end_tstamp=None, email_sent=False, comment=comment
                )
                session.add(batch)
        return begin_tstamp

    async def is_open(self) -> bool:
        async with Session() as session:
            return await self._is_open(session)

    async def get_open(self) -> Batch:
        """Used by the persistent controller"""
        async with Session() as session:
            return await self._get_open(session)

    async def latest(self, load_summaries: bool = False) -> Batch | None:
        """Used by the persistent controller"""
        async with Session() as session:
            return await self._latest(session, load_summaries)

    async def by_date(self, tstamp: datetime, load_summaries: bool = False) -> Batch | None:
        async with Session() as session:
            return await self._by_date(session, tstamp)

    async def close(self) -> Tuple[datetime, datetime, int]:
        end_tstamp = datetime.now(timezone.utc).replace(microsecond=0)
        async with Session() as session:
            async with session.begin():
                await self._assert_open(session)
                batch = await self._get_open(session)
                t0 = batch.begin_tstamp
                t1 = end_tstamp
                # We count summaries even if the upd_flag is False
                q = (
                    select(func.count("*"))
                    .select_from(SummaryView)
                    .where(SummaryView.session.between(t0, t1))
                )
                N = (await session.scalars(q)).one()
                batch.end_tstamp = end_tstamp
                batch.email_sent = False
                batch.calibrations = N
                session.add(batch)
        return t0, t1, N

    async def purge(self) -> int:
        async with Session() as session:
            async with session.begin():
                stmt = delete(Batch).where(Batch.calibrations == 0, Batch.end_tstamp.is_not(None))
                result = await session.execute(stmt)
                return result.rowcount

    async def orphan(self) -> set:
        in_batches = set()
        all_summaries = set()
        async with Session() as session:
            async with session.begin():
                q = select(Batch.begin_tstamp, Batch.end_tstamp).where(
                    Batch.end_tstamp.is_not(None)
                )
                batches = (await session.execute(q)).all()
                q = select(SummaryView.session)
                all_summaries = set((await session.scalars(q)).all())
                for t0, t1 in batches:
                    q = select(SummaryView.session).where(SummaryView.session.between(t0, t1))
                    summaries = (await session.scalars(q)).all()
                    in_batches.update(set(summaries))
        return all_summaries - in_batches

    async def view(
        self,
    ) -> Iterable[Tuple[datetime, datetime, bool, int, str]]:
        async with Session() as session:
            async with session.begin():
                q = select(
                    Batch.begin_tstamp,
                    Batch.end_tstamp,
                    Batch.calibrations,
                    Batch.email_sent,
                    Batch.comment,
                ).order_by(Batch.begin_tstamp.desc())
                batches = (await session.execute(q)).all()
        return batches

    async def export(self, path: str) -> None:
        pass

    # ----------------
    # Helper functions
    # ----------------

    async def _is_open(
        self,
        session: Session,
    ) -> bool:
        q = select(func.count("*")).select_from(Batch).where(Batch.end_tstamp.is_(None))
        count = (await session.scalars(q)).one()
        return count > 0

    async def _get_open(
        self,
        session: Session,
    ) -> Batch | None:
        q = select(Batch).where(Batch.end_tstamp.is_(None))
        batch = (await session.scalars(q)).one_or_none()
        return batch

    async def _latest(
        self,
        session: Session,
        load_summaries: bool,
    ) -> Batch | None:
        sub_q = select(func.max(Batch.begin_tstamp))
        q = select(Batch).where(Batch.begin_tstamp.in_(sub_q))
        batch = (await session.scalars(q)).one()
        if load_summaries:
            summaries = await batch.awaitable_attrs.summaries
            for summary in summaries:
                await summary.awaitable_attrs.photometer
        return batch

    async def _by_date(
        self,
        session: Session,
        tstamp: datetime,
        load_summaries: bool,
    ) -> Batch | None:
        q = select(Batch).where(Batch.begin_tstamp == tstamp)
        batch = (await session.scalars(q)).one_or_none()
        if batch and load_summaries:
            summaries = await batch.awaitable_attrs.summaries
            for summary in summaries:
                await summary.awaitable_attrs.photometer
        return batch

    async def _assert_closed(
        self,
        session: Session,
    ) -> None:
        already_open = await self._is_open(session)
        if already_open:
            raise RuntimeError("Batch already open!")

    async def _assert_open(
        self,
        session: Session,
    ) -> None:
        already_open = await self._is_open(session)
        if not already_open:
            raise RuntimeError("Batch already closed!")



import re
import argparse


from ...constants import SERIAL_PORT_PREFIX, TEST_SERIAL_PORT, TEST_BAUD
from ...constants import TEST_IP, TEST_TCP_PORT, TEST_UDP_PORT


def valid_ip_address(ip: str):
    """Validate an IPv4 address returning True or False"""
    return [
        0 <= int(x) < 256 for x in re.split(r"\.", re.match(r"^\d+\.\d+\.\d+\.\d+$", ip).group(0))
    ].count(True) == 4


def vendpoint(
    value,
    default_ip: str = TEST_IP,
    default_tcp_port: int = TEST_TCP_PORT,
    default_udp_port: int = TEST_UDP_PORT,
    default_serial_port: str = TEST_SERIAL_PORT,
    default_baud: int = TEST_BAUD,
) -> str:
    """
    Utility to convert command line values to serial or tcp endpoints
    tcp
    tcp::<port>
    tcp:<ip>
    tcp:<ip>:<port>
    udp
    udp::<port>
    udp:<ip>
    udp:<ip>:<port>
    serial
    serial::<baud>
    serial:<serial_port>
    serial:<serial_port>:<baud>

    """
    parts = [elem.strip() for elem in value.split(":")]
    length = len(parts)
    if length < 1 or length > 3:
        raise argparse.ArgumentTypeError("Invalid endpoint format {0}".format(value))
    proto = parts[0]
    if proto == "tcp":
        if length == 1:
            ip = str(default_ip)
            port = str(default_tcp_port)
        elif length == 2:
            ip = parts[1]
            port = str(default_tcp_port)
        elif valid_ip_address(parts[1]):
            ip = parts[1]
            port = parts[2]
        else:
            ip = str(default_ip)
            port = parts[2]
        result = proto + ":" + ip + ":" + port
    elif proto == "serial":
        if length == 1:
            serial = SERIAL_PORT_PREFIX + str(default_serial_port)
            baud = str(default_baud)
        elif length == 2:
            serial = SERIAL_PORT_PREFIX + str(parts[1])
            baud = str(default_baud)
        elif parts[1] != "":
            serial = SERIAL_PORT_PREFIX + str(parts[1])
            baud = parts[2]
        else:
            serial = SERIAL_PORT_PREFIX + str(default_serial_port)
            baud = parts[2]
        result = proto + ":" + serial + ":" + baud
    elif proto == "udp":
        if length == 1:
            ip = str(default_ip)
            port = str(default_udp_port)
        elif length == 2:
            ip = parts[1]
            port = str(default_udp_port)
        elif valid_ip_address(parts[1]):
            ip = parts[1]
            port = parts[2]
        else:
            ip = str(default_ip)
            port = parts[2]
        result = proto + ":" + ip + ":" + port
    else:
        raise argparse.ArgumentTypeError("Invalid endpoint prefix {0}".format(parts[0]))
    return result



# ----------------------------------------------------------------------
# Copyright (c) 2024 Rafael Gonzalez.
#
# See the LICENSE file for details
# ----------------------------------------------------------------------

# --------------------
# System wide imports
# -------------------

import logging
from argparse import Namespace, ArgumentParser

# -------------------
# Third party imports
# -------------------

from lica.sqlalchemy import sqa_logging
from lica.asyncio.cli import execute
from lica.asyncio.photometer import Role

# --------------
# local imports
# -------------

from .. import __version__
from ..controller.photometer import Writer
from .util.misc import log_phot_info, update_zp
from .util import parser as prs
from ..dao import engine

# ----------------
# Module constants
# ----------------

DESCRIPTION = "TESS-W Zero Point update tool"

# -----------------------
# Module global variables
# -----------------------

# get the module logger
log = logging.getLogger(__name__.split(".")[-1])
controller = None

# -----------------
# Auxiliary classes
# -----------------


# -------------------
# Auxiliary functions
# -------------------


async def cli_update_zp(args: Namespace) -> None:
    global controller

    test_params = {
        "model": args.test_model,
        "sensor": args.test_sensor,
        "endpoint": args.test_endpoint,
        "old_proto": args.test_old_proto,
        "log_level": logging.DEBUG if args.test_raw_message else logging.INFO,
        "strict": args.test_strict,
    }
    controller = Writer(
        test_params=test_params,
    )
    await controller.init()
    await log_phot_info(controller, Role.TEST)
    if args.dry_run:
        return
    await update_zp(controller, args.zero_point)


# -----------------
# CLI API functions
# -----------------


def add_args(parser: ArgumentParser):
    subparser = parser.add_subparsers(dest="command", required=True)
    p = subparser.add_parser("test", parents=[prs.wrzp(), prs.test()], help="Read test photometer")
    p.add_argument("-d", "--dry-run", action="store_true", help="Do not update photometer")
    p.set_defaults(func=cli_update_zp)


async def cli_main(args: Namespace) -> None:
    sqa_logging(args)
    await args.func(args)
    await engine.dispose()


def main():
    """The main entry point specified by pyproject.toml"""
    execute(
        main_func=cli_main,
        add_args_func=add_args,
        name=__name__,
        version=__version__,
        description=DESCRIPTION,
    )


if __name__ == "__main__":
    main()



# ---------------------------
# Third-party library imports
# ----------------------------

from sqlalchemy import select
from zptessdao.asyncio import Config

# --------------
# local imports
# -------------

from ..dao import Session

async def load_config(session: Session, section: str, prop: str) -> str | None:
    q = select(Config.value).where(Config.section == section, Config.prop == prop)
    return (await session.scalars(q)).one_or_none()


# ----------------------------------------------------------------------
# Copyright (c) 2024 Rafael Gonzalez.
#
# See the LICENSE file for details
# ----------------------------------------------------------------------

# --------------------
# System wide imports
# -------------------


# ---------------------
# Third party libraries
# ---------------------

from lica.misc import chop
from lica.asyncio.photometer import Role, Model
from lica.asyncio.photometer.protocol import UdpProtocol, TcpProtocol, SerialProtocol
from lica.asyncio.photometer.payload import JsonPayload, OldPayload
from lica.asyncio.photometer.photinfo import HTMLInfo, DBaseInfo
from lica.asyncio.photometer.photometer import Photometer


class PhotometerBuilder:
    def __init__(self, engine=None):
        self._engine = engine

    def build(
        self, model: Model, role: Role, endpoint: str | None = None, strict: bool = False
    ) -> Photometer:
        url = role.endpoint() if endpoint is None else endpoint
        transport, name, number = chop(url, sep=":")
        number = int(number) if number else 80
        photometer = Photometer(role)

        if role == Role.REF:
            assert model is Model.TESSW, "Reference photometer model should be TESS-W"
            assert transport == "serial", "Reference photometer should use a serial transport"
            assert self._engine is not None, "Database engine is needed for the REF photometer"
            info_obj = DBaseInfo(logger=photometer.log, engine=self._engine)
            transport_obj = SerialProtocol(logger=photometer.log, port=name, baudrate=number)
            decoder_obj = OldPayload(logger=photometer.log, strict=strict)
        else:
            if transport == "serial":
                # Although we are able to get readings from serial port,
                # there is no PhotInfo object available
                # The DBaseInfo is only available for the reference photometer.
                raise NotImplementedError(
                    "Test Photometer calibration not possible using a serial port"
                )
            elif transport == "tcp":
                assert model is Model.TESSW, "Test photometer using TCP should be a TESS-W model"
                info_obj = HTMLInfo(logger=photometer.log, addr=name)
                transport_obj = TcpProtocol(logger=photometer.log, host=name, port=number)
                decoder_obj = OldPayload(logger=photometer.log, strict=strict)
            elif transport == "udp":
                assert model is Model.TESSW, "Test photometer using UDP should be a TESS-W model"
                info_obj = HTMLInfo(logger=photometer.log, addr=name)
                transport_obj = UdpProtocol(logger=photometer.log, local_port=number)
                decoder_obj = JsonPayload(logger=photometer.log, strict=strict)
            else:
                raise ValueError(f"Transport {transport} not known")
        photometer.attach(transport_obj, info_obj, decoder_obj)
        return photometer



# ----------------------------------------------------------------------
# Copyright (c) 2024 Rafael Gonzalez.
#
# See the LICENSE file for details
# ----------------------------------------------------------------------

# --------------------
# System wide imports
# -------------------

import logging


from datetime import datetime

# ---------------------------
# Third-party library imports
# ----------------------------

from sqlalchemy import select
from lica.asyncio.photometer import Role
from zptessdao.asyncio import SampleView

# --------------
# local imports
# -------------

from ..dao import Session
from ..constants import FreqSequence, TimeSequence, NameSequence

# -----------------------
# Module global variables
# -----------------------

# get the module logger
log = logging.getLogger(__name__.split(".")[-1])


# ------------------
# Auxiliar functions
# ------------------


class Controller:
    def __init__(self):
        pass

    async def samples(self, session_id: datetime, role: Role) -> tuple[FreqSequence, TimeSequence, NameSequence]:
        """Used by the persistent controller"""
        log.info("fetching samples from session %s", session_id)
        async with Session() as session:
            async with session.begin():
                q = (
                    select(SampleView.freq, SampleView.tstamp, SampleView.name).distinct()
                    .where(SampleView.session == session_id, SampleView.role == role)
                )
                samples = (await session.execute(q)).all()
                log.info("found %d %s samples", len(samples), role)
                return zip(*samples)




# ----------------------------------------------------------------------
# Copyright (c) 2024 Rafael Gonzalez.
#
# See the LICENSE file for details
# ----------------------------------------------------------------------

# --------------------
# System wide imports
# -------------------

import json
import logging
import asyncio
import statistics

from dataclasses import dataclass
from collections import defaultdict
from logging import Logger
from argparse import Namespace, ArgumentParser
from datetime import datetime, timezone
from typing import Tuple

# -------------------
# Third party imports
# -------------------


from lica.sqlalchemy import sqa_logging
from lica.asyncio.cli import execute

# --------------
# local imports
# -------------

from .. import __version__
from .util import parser as prs
from ..dao import engine

# ----------------
# Module constants
# ----------------

DESCRIPTION = "TESS-W Multi photometer reader tool"

# -----------------------
# Module global variables
# -----------------------

# get the module logger
log = logging.getLogger(__name__.split(".")[-1])
controller = None

# ------------------
# Auxiliar functions
# ------------------

# -----------------
# Auxiliary classes
# -----------------


class UdpProtocol(asyncio.DatagramProtocol):
    def __init__(
        self,
        logger: Logger,
        local_host: str = "0.0.0.0",
        local_port: int = 2255,
        loop: asyncio.AbstractEventLoop | None = None,
        encoding: str = "utf-8",
        newline: bytes = b"\r\n",
    ):
        self.loop = loop or asyncio.get_event_loop()
        self.log = logger
        self.encoding = encoding
        self.newline = newline
        self.local_host = local_host
        self.local_port = local_port
        # Futures for external awaiters
        self.on_data_received: asyncio.Future | None = None
        self.on_conn_lost: asyncio.Future = self.loop.create_future()
        self.log.info("Using %s", self.__class__.__name__)

    async def open(self) -> None:
        self.log.debug("Opening UDP endpoint on (%s, %s)", self.local_host, self.local_port)
        transport, self.protocol = await self.loop.create_datagram_endpoint(
            lambda: self, local_addr=(self.local_host, self.local_port)
        )

    def close(self) -> None:
        self.log.debug("Closing %s transport", self.transport.__class__.__name__)
        self.transport.close()

    # ---------------------------------------
    # The asyncio Protocol callback interface
    # ---------------------------------------

    def connection_made(self, transport: asyncio.Transport) -> None:
        self.log.debug("UDP socket listening to (%s, %s)", self.local_host, self.local_port)
        self.transport = transport

    def connection_lost(self, exc: Exception | None) -> None:
        self.log.debug("Closed UDP endpoint on (%s, %s)", self.local_host, self.local_port)
        if not self.on_conn_lost.cancelled() and not self.on_conn_lost.done():
            self.on_conn_lost.set_result(True)
        if (
            self.on_data_received is not None
            and not self.on_data_received.done()
            and not self.on_data_received.cancelled()
        ):
            self.on_data_received.set_exception(
                ConnectionError("UDP socket closed before incoming message was complete")
            )
        self.transport.close()

    def datagram_received(self, payload: bytes, addr: str):
        now = datetime.now(timezone.utc)
        message = payload.decode(self.encoding, errors="replace")
        try:
            message = json.loads(message)
        except Exception as e:
            self.log.exception(e)
        else:
            if isinstance(message, dict):
                message["seq"] = message["udp"]
                del message["udp"]
                if (
                    self.on_data_received is not None
                    and not self.on_data_received.cancelled()
                    and not self.on_data_received.done()
                ):
                    self.on_data_received.set_result((now, message))

    # ----------------------
    # The iterator interface
    # ----------------------

    def __aiter__(self) -> "UdpProtocol":
        # The iterator is its own async iterator.
        return self

    async def __anext__(self) -> Tuple[datetime, str]:
        if self.on_data_received is not None and not self.on_data_received.done():
            self.on_data_received.cancel()
        self.on_data_received = self.loop.create_future()
        return await self.on_data_received


@dataclass
class Stats:
    N: int
    mean: float
    median: float
    stdev: float
    mode: float

    def __repr__(self) -> str:
        return f"N={self.N:03d}, mean={self.mean:.2f}, \u03c3={self.stdev:.3f}, median={self.median:.2f}, mode={self.mode:.2f}"


# -------------------
# Auxiliary functions
# -------------------


def stats_by_name(hashable) -> dict[str, Stats]:
    stats = dict()
    for name, messages in hashable.items():
        values = [msg["mag"] for msg in messages]
        mean = statistics.fmean(values)
        median = statistics.median_low(values)
        stdev = statistics.stdev(values, xbar=mean)
        modes = statistics.multimode(values)
        stats[name] = Stats(N=len(values), mean=mean, median=median, stdev=stdev, mode=modes[0])
    return stats


async def cli_multi(args: Namespace) -> None:
    meas_session = datetime.now(timezone.utc)
    log.info("Session de medidas %s", meas_session.strftime("%Y-%m-%dT%H:%M:%S"))
    proto = UdpProtocol(log)
    N = args.num_messages
    messages = defaultdict(list)
    await proto.open()
    for i in range(N):
        tstamp, msg = await anext(proto)
        log.info("%s %s",tstamp.strftime("%H:%M:%S.%f"), msg)
        msg["tstamp"] = tstamp
        messages[msg["name"]].append(msg)
    stats = stats_by_name(messages)
    for name, stat in stats.items():
        log.info("%8s => %s", name, stat)


# -----------------
# CLI API functions
# -----------------


def add_args(parser: ArgumentParser):
    subparser = parser.add_subparsers(dest="command", required=True)
    p = subparser.add_parser(
        "test",
        parents=[prs.nmsg()],
        help="Read several test photometers",
    )
    p.set_defaults(func=cli_multi)


async def cli_main(args: Namespace) -> None:
    sqa_logging(args)
    await args.func(args)
    await engine.dispose()


def main():
    """The main entry point specified by pyproject.toml"""
    execute(
        main_func=cli_main,
        add_args_func=add_args,
        name=__name__,
        version=__version__,
        description=DESCRIPTION,
    )


if __name__ == "__main__":
    main()



# Re-export classes

from .base import Controller
from .reader import Controller as Reader
from .writer import Controller as Writer
from .volatile import Controller as VolatileCalibrator
from .persistent import Controller as PersistentCalibrator
from .types import Event, RoundStatistics, RoundStatsType

__all__ = [
    "Controller",
    "Reader",
    "Writer",
    "VolatileCalibrator",
    "PersistentCalibrator",
    "Event",
    "RoundStatistics",
    "RoundStatsType",
]



# ----------------------------------------------------------------------
# Copyright (c) 2024 Rafael Gonzalez.
#
# See the LICENSE file for details
# ----------------------------------------------------------------------

# --------------------
# System wide imports
# -------------------

import os
import asyncio
import logging

from argparse import Namespace, ArgumentParser

# -------------------
# Third party imports
# -------------------

from lica.sqlalchemy import sqa_logging
from lica.asyncio.cli import execute
from lica.tabulate import paging

# --------------
# local imports
# -------------

from .. import __version__
from .util import parser as prs

from ..controller.batch import Controller as BatchController
from ..controller.exporter import Controller as Exporter
from ..dao import engine

# ----------------
# Module constants
# ----------------

TSTAMP_FMT = "%Y-%m-%dT%H:%M:%S"

# -----------------------
# Module global variables
# -----------------------

# get the module logger
log = logging.getLogger(__name__.split(".")[-1])


# -----------------
# Auxiliar function
# -----------------

# -----------------
# CLI API functions
# -----------------


async def cli_batch_begin(args: Namespace) -> None:
    batch = BatchController()
    tstamp = await batch.open(comment="pepe")
    log.info("Opening batch %s", tstamp.strftime(TSTAMP_FMT))


async def cli_batch_end(args: Namespace) -> None:
    batch = BatchController()
    t0, t1, N = await batch.close()
    log.info(
        "Closing batch [%s - %s] with %d calibrations",
        t0.strftime(TSTAMP_FMT),
        t1.strftime(TSTAMP_FMT),
        N,
    )


async def cli_batch_purge(args: Namespace) -> None:
    batch = BatchController()
    N = await batch.purge()
    log.info("Purged %d batches with no summary calibration entries", N)


async def cli_batch_orphan(args: Namespace) -> None:
    batch = BatchController()
    orphans = await batch.orphan()
    log.info("%d orphan summaries not belonging to a batch", len(orphans))
    if args.list:
        for i, item in enumerate(sorted(orphans), start=1):
            log.info("[%03d] %s", i, item)


async def cli_batch_view(args: Namespace) -> None:
    batch = BatchController()
    HEADERS = ("Begin (UTC)", "End (UTC)", "# Sessions", "Emailed?", "Comment")
    iterable = await batch.view()
    paging(iterable, HEADERS, page_size=args.page_size, table_fmt=args.table_format)


async def cli_batch_export(args: Namespace) -> None:
    if args.all:
        exporter = Exporter(base_dir=args.base_dir, filename_prefix="all")
        log.info("exporting to directory %s", args.base_dir)
        summaries = await exporter.query_summaries()
        # This is not necessary in a CLI application, just to show how it should be done in a GUI
        await asyncio.to_thread(exporter.export_summaries, summaries)
    else:
        batch_ctrl = BatchController()
        batch = (
            await batch_ctrl.by_date(args.begin_date)
            if args.begin_date
            else await batch_ctrl.latest()
        )
        if batch is not None:
            t0 = batch.begin_tstamp.strftime("%Y%m%d")
            t1 = batch.end_tstamp.strftime("%Y%m%d")
            filename_preffix = f"from_{t0}_to_{t1}"
            base_dir = os.path.join(args.base_dir, filename_preffix)
            log.info("exporting to directory %s", base_dir)
            os.makedirs(base_dir, exist_ok=True)
            exporter = Exporter(
                base_dir=base_dir,
                filename_prefix=filename_preffix,
                begin_tstamp=batch.begin_tstamp,
                end_tstamp=batch.end_tstamp,
            )
            summaries = await exporter.query_summaries()
            await asyncio.to_thread(exporter.export_summaries, summaries)
            rounds = await exporter.query_rounds()
            await asyncio.to_thread(exporter.export_rounds, rounds)
            samples = await exporter.query_samples()
            await asyncio.to_thread(exporter.export_samples, samples)
            zip_file_path = await asyncio.to_thread(exporter.pack)
            if not args.email:
                log.info("Not sending email for this batch")
                return
            if batch.email_sent is None:
                log.info("Never tried to send an email for this batch")
            elif batch.email_sent == 0:
                log.info("Tried to previously send email for this batch, but failed")
            else:
                log.info("Already sent an email for this batch")
                return
            await exporter.load_email_config()
            internet = await exporter.check_internet()
            if not internet:
                log.error("No connection to Internet. Stopping here")
                return
            email_sent = await asyncio.to_thread(exporter.send_email, zip_file_path)
            await exporter.update_batch(batch, email_sent)
        else:
            log.info("No batch is available")


def add_args(parser: ArgumentParser):
    subparser = parser.add_subparsers(dest="command", required=True)
    p = subparser.add_parser(
        "begin",
        parents=[prs.comm()],
        help="Begin new calibration batch",
    )
    p.set_defaults(func=cli_batch_begin)
    p = subparser.add_parser(
        "end",
        parents=[],
        help="End current calibration batch",
    )
    p.set_defaults(func=cli_batch_end)
    p = subparser.add_parser(
        "purge",
        parents=[],
        help="Purge empty calibration batches",
    )
    p.set_defaults(func=cli_batch_purge)
    p = subparser.add_parser(
        "view",
        parents=[prs.tbl()],
        help="List calibration batches",
    )
    p.set_defaults(func=cli_batch_view)
    p = subparser.add_parser(
        "orphan",
        parents=[prs.lst()],
        help="List calibration mummaries not belonging to any batch",
    )
    p.set_defaults(func=cli_batch_orphan)
    p = subparser.add_parser(
        "export",
        parents=[prs.odir(), prs.expor()],
        help="Export calibration batch to CSV files",
    )
    p.set_defaults(func=cli_batch_export)


async def cli_main(args: Namespace) -> None:
    sqa_logging(args)
    await args.func(args)
    await engine.dispose()


def main():
    """The main entry point specified by pyproject.toml"""
    execute(
        main_func=cli_main,
        add_args_func=add_args,
        name=__name__,
        version=__version__,
        description="Batch calibration management tools",
    )



# ----------------------------------------------------------------------
# Copyright (c) 2024 Rafael Gonzalez.
#
# See the LICENSE file for details
# ----------------------------------------------------------------------

# --------------------
# System wide imports
# -------------------

import math
import datetime
import logging
import asyncio
import statistics

from collections import defaultdict
from typing import Any, Mapping, Sequence


# ---------------------------
# Third-party library imports
# ----------------------------


from pubsub import pub
from lica.asyncio.photometer import Role
from zptessdao.constants import CentralTendency

# --------------
# local imports
# -------------

from .util import best
from .types import Event, RoundStatistics, SummaryStatistics
from .ring import RingBuffer, Message
from .base import Controller as BaseController
from .. import load_config
from ...dao import Session


# ----------------
# Module constants
# ----------------


SECTION = {Role.REF: "ref-stats", Role.TEST: "test-stats"}

# -----------------------
# Module global variables
# -----------------------

# get the module logger
log = logging.getLogger(__name__.split(".")[-1])

# -------------------
# Auxiliary functions
# -------------------

# -----------------
# Auxiliary classes
# -----------------


class Controller(BaseController):
    """
    In-memory Photometer Calibration Controller
    """

    def __init__(
        self,
        ref_params: Mapping[str, Any] | None = None,
        test_params: Mapping[str, Any] | None = None,
        common_params: Mapping[str, Any] | None = None,
    ):
        super().__init__(ref_params, test_params)
        self.common_param = common_params
        self.period = None
        self.central = None
        self.nrounds = None
        self.zp_fict = None
        self.zp_offset = None
        self.zp_abs = None
        self.author = None
        self.accum_samples = defaultdict(list)
        self.time_intervals = defaultdict(list)
        self._unique_samples = defaultdict(set)

    # ==========
    # Public API
    # ==========

    async def init(self) -> None:
        await super().init()
        self.meas_session = datetime.datetime.now(datetime.timezone.utc).replace(microsecond=0)
        async with Session() as session:
            val_db = await load_config(session, SECTION[Role.TEST], "samples")
            val_arg = self.common_param["buffer"]
            self.capacity = val_arg if val_arg is not None else int(val_db)
            val_db = await load_config(session, SECTION[Role.TEST], "period")
            val_arg = self.common_param["period"]
            self.period = val_arg if val_arg is not None else float(val_db)
            val_db = await load_config(session, SECTION[Role.TEST], "central")
            val_arg = self.common_param["central"]
            self.central = val_arg if val_arg is not None else CentralTendency(val_db)
            val_db = await load_config(session, "calibration", "zp_fict")
            val_arg = self.common_param["zp_fict"]
            self.zp_fict = val_arg if val_arg is not None else float(val_db)
            val_db = await load_config(session, "calibration", "rounds")
            val_arg = self.common_param["rounds"]
            self.nrounds = val_arg if val_arg is not None else int(val_db)
            val_db = await load_config(session, "calibration", "offset")
            val_arg = self.common_param["zp_offset"]
            self.zp_offset = val_arg if val_arg is not None else float(val_db)
            val_db = await load_config(session, "calibration", "author")
            val_arg = self.common_param["author"]
            self.author = val_arg if val_arg is not None else val_db
            # The absolute ZP is the stored ZP in the reference photometer.
            self.zp_abs = float(await load_config(session, "ref-device", "zp"))
        self.persist = self.common_param["persist"]
        self.update = self.common_param["update"]
        for role in self.roles:
            self.ring[role] = RingBuffer(capacity=self.capacity, central=self.central)

    async def calibrate(self) -> float:
        """
        Calibrate the Test photometer against the Reference Photometer
        and return the final Zero Point to Write to the Test Photometer
        """
        self._on_calib_start()
        # Waiting for both circular buffers to be filled
        try:
            async with asyncio.TaskGroup() as tg:
                for role in self.roles:
                    tg.create_task(self._fill_buffer_task(role))
        except* Exception as eg:
            log.error(eg.exceptions)
        # launch the background buffer filling task and the stats task
        try:
            self.is_calibrated = False
            async with asyncio.TaskGroup() as tg:
                for role in self.roles:
                    tg.create_task(self._producer_task(role))
                stat_task = tg.create_task(self._statistics())
        except* Exception as eg:
            log.error(eg.exceptions)
        zero_points, freqs = stat_task.result()
        final_zero_point = self._post_statistics(zero_points, freqs)
        self._on_calib_end()
        return final_zero_point

    async def not_updated(self, zero_point: float, msg: str):
        pass

    def unique_samples(self, role: Role) -> set[Message]:
        return self._unique_samples[role]

    # ===========
    # Private API
    # ===========

    async def _fill_buffer_task(self, role: Role) -> None:
        """Finite task to fill the ring buffer"""
        async with self.photometer[role]:
            while len(self.ring[role]) < self.capacity:
                msg = await anext(self.photometer[role].readings)
                if msg is not None:
                    self.ring[role].append(msg)
                    pub.sendMessage(Event.READING, role=role, reading=msg)

    async def _producer_task(self, role: Role) -> None:
        """This task continues to re-fill the buffer when statistics are being computed"""
        async with self.photometer[role]:
            while not self.is_calibrated:
                msg = await anext(self.photometer[role].readings)
                if msg is not None:
                    self.ring[role].append(msg)

    def _magnitude(self, role: Role, freq: float, freq_offset):
        return self.zp_fict - 2.5 * math.log10(freq - freq_offset)

    # --------------------
    # Hooks implementation
    # --------------------

    def _on_calib_start(self) -> None:
        pub.sendMessage(Event.CAL_START)

    def _on_calib_end(self) -> None:
        pub.sendMessage(Event.CAL_END)

    def _on_round(self, round_info: Mapping[str, Any]) -> None:
        pub.sendMessage(Event.ROUND, **round_info)

    def _on_summary(self, summary_info: Mapping[str, Any]) -> None:
        pub.sendMessage(Event.SUMMARY, **summary_info)

    # ----------------------
    # Private helper methods
    # ----------------------

    def _round_statistics(self, role: Role) -> RoundStatistics:
        log = logging.getLogger(role.tag())
        freq_offset = self.phot_info[role]["freq_offset"]
        freq = stdev = mag = None
        try:
            freq, stdev = self.ring[role].statistics()
            mag = self._magnitude(role, freq, freq_offset)
        except statistics.StatisticsError as e:
            log.error("Statistics error: %s", e)
        except ValueError as e:
            log.error("math.log10() error for freq=%s, freq_offset=%s}: %s", freq, freq_offset, e)
        finally:
            return freq, stdev, mag

    async def _statistics(self) -> SummaryStatistics:
        zero_points = list()
        stats = list()
        freqs = dict()
        for i in range(0, self.nrounds):
            stats_per_round = dict()
            for role in self.roles:
                stats_per_round[role] = self._round_statistics(role)
                samples = self.ring[role].copy()
                self.accum_samples[role].append(samples)
                self._unique_samples[role].update(samples)
                self.time_intervals[role].append(self.ring[role].intervals())
            mag_diff = stats_per_round[Role.REF][2] - stats_per_round[Role.TEST][2]
            zero_points.append(self.zp_abs + mag_diff)
            stats.append(stats_per_round)
            round_info = {
                "current": i + 1,
                "mag_diff": mag_diff,
                "zero_point": zero_points[i],
                "stats": stats_per_round,
            }
            self._on_round(round_info)
            if i != self.nrounds - 1:
                await asyncio.sleep(self.period)
        zero_points = [round(zp, 2) for zp in zero_points]
        for role in self.roles:
            freqs[role] = [stats_pr[role][0] for stats_pr in stats]
        self.is_calibrated = True  # So no more buffer filling
        return zero_points, freqs

    def _overlapping_windows(self) -> Mapping[Role, Sequence[float | None]]:
        overlaps = defaultdict(list)
        for role in self.roles:
            for i, t in enumerate(self.time_intervals[role]):
                if i < (self.nrounds - 1):
                    next_t = self.time_intervals[role][i + 1]
                    T = (t[1] - next_t[0]).total_seconds()
                    T = None if T <= 0 else T
                    overlaps[role].append(T)
        return overlaps

    def _post_statistics(self, zero_points, freqs) -> float:
        best_zp_method, best_zero_point = best(zero_points)
        best_freq = dict()
        best_freq_method = dict()
        best_mag = dict()
        for role in self.roles:
            best_freq_method[role], best_freq[role] = best(freqs[role])
            best_mag[role] = self.zp_fict - 2.5 * math.log10(best_freq[role])
        final_zero_point = best_zero_point + self.zp_offset
        mag_diff = -2.5 * math.log10(best_freq[Role.REF] / best_freq[Role.TEST])
        overlap = self._overlapping_windows()
        summary_info = {
            "zero_point_seq": zero_points,
            "freq_seq": freqs,
            "best_freq": best_freq,
            "best_freq_method": best_freq_method,
            "best_mag": best_mag,
            "mag_diff": mag_diff,
            "best_zero_point": best_zero_point,
            "best_zero_point_method": best_zp_method,
            "final_zero_point": final_zero_point,
            "overlapping_windows": overlap,
        }
        self._on_summary(summary_info)
        return final_zero_point



# ----------------------------------------------------------------------
# Copyright (c) 2024 Rafael Gonzalez.
#
# See the LICENSE file for details
# ----------------------------------------------------------------------

# --------------------
# System wide imports
# -------------------

import sys
import logging
import asyncio
import statistics
from argparse import Namespace, ArgumentParser
from datetime import datetime, timezone

# -------------------
# Third party imports
# -------------------

import aiohttp

from lica.sqlalchemy import sqa_logging
from lica.asyncio.cli import execute
from lica.asyncio.photometer import Role

# --------------
# local imports
# -------------

from .. import __version__
from ..controller.photometer import Reader
from .util import parser as prs
from .util.misc import log_phot_info, log_messages, log_and_exit
from ..dao import engine
from ..mpl import plot

# ----------------
# Module constants
# ----------------

DESCRIPTION = "TESS-W Reader tool"

# -----------------------
# Module global variables
# -----------------------

# get the module logger
log = logging.getLogger(__name__.split(".")[-1])
controller = None

# ------------------
# Auxiliar functions
# ------------------

# -----------------
# Auxiliary classes
# -----------------


# -------------------
# Auxiliary functions
# -------------------


async def cli_read_ref(args: Namespace) -> None:
    global controller
    meas_session = datetime.now(timezone.utc)
    ref_params = {
        "model": args.ref_model,
        "sensor": args.ref_sensor,
        "endpoint": args.ref_endpoint,
        "old_proto": args.ref_old_proto,
        "log_level": logging.DEBUG if args.ref_raw_message else logging.INFO,
        "strict": args.ref_strict,
    }
    controller = Reader(
        ref_params=ref_params,
    )
    try:
        await controller.init()
        await log_phot_info(controller, Role.REF)
        if not args.info:
            messages = await log_messages(controller, Role.REF, args.num_messages)
            if args.plot_histo:
                freqs = [msg["freq"] for msg in messages]
                decimals = 2 if statistics.mean(freqs) > 3 else 3
                tstamps = [msg["tstamp"] for msg in messages]
                name = controller.phot_info[Role.REF]["name"]
                plot.histograms(
                    session=meas_session,
                    roles=[Role.REF],
                    freqs=[freqs],
                    tstamps=[tstamps],
                    names=[name],
                    decimals=[decimals],
                )
    except asyncio.TimeoutError as e:
        log_and_exit(log, e, args.trace)
    except aiohttp.client_exceptions.ClientConnectorError as e:
        log_and_exit(log, e, args.trace)



async def cli_read_test(args: Namespace) -> None:
    global controller
    meas_session = datetime.now(timezone.utc)
    test_params = {
        "model": args.test_model,
        "sensor": args.test_sensor,
        "endpoint": args.test_endpoint,
        "old_proto": args.test_old_proto,
        "log_level": logging.DEBUG if args.test_raw_message else logging.INFO,
        "strict": args.test_strict,
    }
    controller = Reader(
        test_params=test_params,
    )
    try:
        await controller.init()
        await log_phot_info(controller, Role.TEST)
        if not args.info:
            messages = await log_messages(controller, Role.TEST, args.num_messages)
            if args.plot_histo:
                freqs = [msg["freq"] for msg in messages]
                tstamps = [msg["tstamp"] for msg in messages]
                name = controller.phot_info[Role.TEST]["name"]
                plot.histograms(
                    session=meas_session,
                    roles=[Role.REF],
                    freqs=[freqs],
                    tstamps=[tstamps],
                    names=[name],
                    decimals=[2],
                )
    except asyncio.TimeoutError as e:
        log_and_exit(log, e, args.trace)
    except aiohttp.client_exceptions.ClientConnectorError as e:
        log_and_exit(log, e, args.trace)



async def cli_read_both(args: Namespace) -> None:
    global controller
    meas_session = datetime.now(timezone.utc)
    ref_params = {
        "model": args.ref_model,
        "sensor": args.ref_sensor,
        "endpoint": args.ref_endpoint,
        "old_proto": args.ref_old_proto,
        "log_level": logging.DEBUG if args.ref_raw_message else logging.INFO,
        "strict": args.ref_strict,
    }
    test_params = {
        "model": args.test_model,
        "sensor": args.test_sensor,
        "endpoint": args.test_endpoint,
        "old_proto": args.test_old_proto,
        "log_level": logging.DEBUG if args.test_raw_message else logging.INFO,
        "strict": args.test_strict,
    }
    controller = Reader(
        ref_params=ref_params,
        test_params=test_params,
    )
    try:
        await controller.init()
        async with asyncio.TaskGroup() as tg:
            tg.create_task(log_phot_info(controller, Role.REF))
            tg.create_task(log_phot_info(controller, Role.TEST))
        if args.info:
            return
        async with asyncio.TaskGroup() as tg:
            task_ref = tg.create_task(log_messages(controller, Role.REF, args.num_messages))
            task_tst = tg.create_task(log_messages(controller, Role.TEST, args.num_messages))
        ref_freqs = [msg["freq"] for msg in task_ref.result()]
        ref_tstamps = [msg["tstamp"] for msg in task_ref.result()]
        tst_freqs = [msg["freq"] for msg in task_tst.result()]
        tst_tstamps = [msg["tstamp"] for msg in task_tst.result()]
        ref_name = controller.phot_info[Role.REF]["name"]
        tst_name = controller.phot_info[Role.TEST]["name"]
        decimals = 2 if statistics.mean(ref_freqs) > 3 else 3
        if args.plot_histo:
            plot.histograms(
                session=meas_session,
                roles=[Role.REF, Role.TEST],
                freqs=[ref_freqs, tst_freqs],
                tstamps=[ref_tstamps, tst_tstamps],
                names=[ref_name, tst_name],
                decimals=[decimals, 2],
            )
        elif args.plot_samples:
            plot.samples(
                session=meas_session,
                roles=[Role.REF, Role.TEST],
                freqs=[ref_freqs, tst_freqs],
                tstamps=[ref_tstamps, tst_tstamps],
                names=[ref_name, tst_name],
            )
        elif args.plot_both:
            plot.histograms(
                session=meas_session,
                roles=[Role.REF, Role.TEST],
                freqs=[ref_freqs, tst_freqs],
                tstamps=[ref_tstamps, tst_tstamps],
                names=[ref_name, tst_name],
                decimals=[decimals, 2],
            )
            plot.samples(
                session=meas_session,
                roles=[Role.REF, Role.TEST],
                freqs=[ref_freqs, tst_freqs],
                tstamps=[ref_tstamps, tst_tstamps],
                names=[ref_name, tst_name],
            )

    except* Exception as eg:
        for e in eg.exceptions:
            if args.trace:
                log.exception(e)
            else:
                log.error(e)
        sys.exit(1)


# -----------------
# CLI API functions
# -----------------


def add_args(parser: ArgumentParser):
    subparser = parser.add_subparsers(dest="command", required=True)
    p = subparser.add_parser(
        "ref",
        parents=[prs.info(), prs.nmsg(), prs.ref(), prs.ploto()],
        help="Read reference photometer",
    )
    p.set_defaults(func=cli_read_ref)
    p = subparser.add_parser(
        "test",
        parents=[prs.info(), prs.nmsg(), prs.test(), prs.ploto()],
        help="Read test photometer",
    )
    p.set_defaults(func=cli_read_test)
    p = subparser.add_parser(
        "both",
        parents=[prs.info(), prs.nmsg(), prs.ref(), prs.test(), prs.ploto()],
        help="read both photometers",
    )
    p.set_defaults(func=cli_read_both)


async def cli_main(args: Namespace) -> None:
    sqa_logging(args)
    await args.func(args)
    await engine.dispose()


def main():
    """The main entry point specified by pyproject.toml"""
    execute(
        main_func=cli_main,
        add_args_func=add_args,
        name=__name__,
        version=__version__,
        description=DESCRIPTION,
    )


if __name__ == "__main__":
    main()



# ----------------------------------------------------------------------
# Copyright (c) 2024 Rafael Gonzalez.
#
# See the LICENSE file for details
# ----------------------------------------------------------------------

# --------------------
# System wide imports
# -------------------

import os

from argparse import ArgumentParser

# ---------------------------
# Third-party library imports
# ----------------------------

from lica.validators import vdir, vdate
from lica.asyncio.photometer import Model as PhotModel, Sensor
from zptessdao.constants import CentralTendency, Calibration


# --------------
# local imports
# -------------

from .validator import vendpoint


def bdir() -> ArgumentParser:
    parser = ArgumentParser(add_help=False)
    parser.add_argument(
        "-b",
        "--base-dir",
        type=vdir,
        default=os.getcwd(),
        metavar="<Dir>",
        help="Base directory for CSV export (default %(default)s)",
    )
    return parser


def idir() -> ArgumentParser:
    parser = ArgumentParser(add_help=False)
    parser.add_argument(
        "-i",
        "--input-dir",
        type=vdir,
        default=os.getcwd(),
        metavar="<Dir>",
        help="Input CSV directory (default %(default)s)",
    )
    return parser


def odir() -> ArgumentParser:
    parser = ArgumentParser(add_help=False)
    parser.add_argument(
        "-o",
        "--output-dir",
        type=vdir,
        default=os.getcwd(),
        metavar="<Dir>",
        help="Output CSV directory (default %(default)s)",
    )
    return parser


def buf() -> ArgumentParser:
    parser = ArgumentParser(add_help=False)
    parser.add_argument(
        "-b",
        "--buffer",
        type=int,
        default=None,
        help="Circular buffer size (default %(default)s)",
    )
    return parser


def info() -> ArgumentParser:
    parser = ArgumentParser(add_help=False)
    parser.add_argument(
        "--info",
        default=False,
        action="store_true",
        help="Query photometer info and exit (default %(default)s)",
    )
    return parser


def persist() -> ArgumentParser:
    parser = ArgumentParser(add_help=False)
    parser.add_argument(
        "-p",
        "--persist",
        default=False,
        action="store_true",
        help="Store calibration results in database (default %(default)s)",
    )
    return parser


def author() -> ArgumentParser:
    parser = ArgumentParser(add_help=False)
    parser.add_argument(
        "-a",
        "--author",
        nargs="+",
        default=None,
        help="Calibration author (default %(default)s)",
    )
    return parser


def upd() -> ArgumentParser:
    parser = ArgumentParser(add_help=False)
    parser.add_argument(
        "-u",
        "--update",
        default=False,
        action="store_true",
        help="Update Zero Point in test photometer (default %(default)s)",
    )
    return parser


def wrzp() -> ArgumentParser:
    parser = ArgumentParser(add_help=False)
    parser.add_argument(
        "-z",
        "--zero-point",
        type=float,
        metavar="<ZP>",
        required=True,
        help="Writed Zero Point to test photometer (default %(default)s)",
    )
    return parser


def nmsg() -> ArgumentParser:
    parser = ArgumentParser(add_help=False)
    parser.add_argument(
        "-N",
        "--num-messages",
        type=int,
        metavar="<N>",
        default=None,
        help="Number of messages to receive (default %(default)s)",
    )
    return parser


def ref() -> ArgumentParser:
    """Reference parser options"""
    parser = ArgumentParser(add_help=False)
    parser.add_argument(
        "-rE",
        "--ref-endpoint",
        type=vendpoint,
        default=None,
        metavar="<ref endpoint>",
        help="Reference photometer endpoint",
    )
    parser.add_argument(
        "-rM",
        "--ref-model",
        type=PhotModel,
        default=None,
        choices=PhotModel,
        help="Ref. photometer model, defaults to %(default)s",
    )
    parser.add_argument(
        "-rO",
        "--ref-old-proto",
        action="store_true",
        default=False,
        help="Use very old protocol instead of JSON, defaults to %(default)s",
    )
    parser.add_argument(
        "-rS",
        "--ref-sensor",
        type=Sensor,
        default=None,
        choices=Sensor,
        help="Reference phot sensor, defaults to %(default)s",
    )
    parser.add_argument(
        "-rR",
        "--ref-raw-message",
        action="store_true",
        default=False,
        help="Log raw messages, defaults to %(default)s",
    )
    parser.add_argument(
        "-rT",
        "--ref-strict",
        action="store_true",
        default=False,
        help="Strict samples rejection by timestamp difference, defaults to %(default)s",
    )
    return parser


def test() -> ArgumentParser:
    """Test parser options"""
    parser = ArgumentParser(add_help=False)
    parser.add_argument(
        "-tE",
        "--test-endpoint",
        type=vendpoint,
        default=None,
        metavar="<test endpoint>",
        help="Test photometer endpoint",
    )
    parser.add_argument(
        "-tM",
        "--test-model",
        type=PhotModel,
        default=None,
        choices=PhotModel,
        help="Test photometer model, defaults to %(default)s",
    )
    parser.add_argument(
        "-tO",
        "--test-old-proto",
        action="store_true",
        default=None,
        help="Use very old protocol instead of JSON, defaults to %(default)s",
    )
    parser.add_argument(
        "-tS",
        "--test-sensor",
        type=Sensor,
        default=None,
        choices=Sensor,
        help="Test photometer sensor, defaults to %(default)s",
    )
    parser.add_argument(
        "-tR",
        "--test-raw-message",
        action="store_true",
        default=False,
        help="Log raw messages, defaults to %(default)s",
    )
    parser.add_argument(
        "-tT",
        "--test-strict",
        action="store_true",
        default=False,
        help="Strict samples rejection by timestamp difference, defaults to %(default)s",
    )
    return parser


def stats() -> ArgumentParser:
    """Statistics parser options"""
    parser = ArgumentParser(add_help=False)
    parser.add_argument(
        "-C",
        "--central",
        type=CentralTendency,
        default=None,
        choices=CentralTendency,
        help="central tendency estimator, defaults to %(default)s",
    )
    parser.add_argument(
        "-R",
        "--rounds",
        type=int,
        default=None,
        metavar="<N>",
        help="Number of calibration rounds, defaults to %(default)s",
    )
    parser.add_argument(
        "-P",
        "--period",
        type=float,
        default=None,
        metavar="<sec.>",
        help="Wait period between rounds[s], defaults to %(default)s",
    )
    parser.add_argument(
        "-Z",
        "--zp-fict",
        type=float,
        default=None,
        metavar="<float>",
        help="Alternative ficticious Zero Point to use for both photometers, defaults to %(default)s",
    )
    parser.add_argument(
        "-O",
        "--zp-offset",
        type=float,
        default=None,
        metavar="<float>",
        help="Offset to add to calibrated Zero Point, defaults to %(default)s",
    )
    return parser


def no_bat() -> ArgumentParser:
    parser = ArgumentParser(add_help=False)
    parser.add_argument(
        "-n",
        "--no-batch",
        default=False,
        action="store_true",
        help="Persistent calibration without an open batch (default %(default)s)",
    )
    return parser


def sess() -> ArgumentParser:
    parser = ArgumentParser(add_help=False)
    parser.add_argument(
        "-s",
        "--session",
        type=vdate,
        metavar="<YYYY-MM-DDTHH:MM:SS>",
        required=True,
        help="session identifier",
    )
    return parser


# ------------------------------
# These are for batch management
# ------------------------------


def comm() -> ArgumentParser:
    parser = ArgumentParser(add_help=False)
    parser.add_argument(
        "-c",
        "--comment",
        type=str,
        nargs="+",
        default=None,
        help="Optional batch comment (default %(default)s)",
    )
    return parser


def tbl() -> ArgumentParser:
    parser = ArgumentParser(add_help=False)
    parser.add_argument(
        "--page-size",
        type=int,
        default=10,
        help="Table page size",
    )
    parser.add_argument(
        "--table-format",
        choices=("simple", "grid"),
        default="simple",
        help="List batches",
    )
    return parser


def lst() -> ArgumentParser:
    parser = ArgumentParser(add_help=False)
    parser.add_argument(
        "--list",
        action="store_true",
        help="List orphan summaries one by one",
    )
    return parser


def expor() -> ArgumentParser:
    parser = ArgumentParser(add_help=False)
    ex1 = parser.add_mutually_exclusive_group(required=True)
    ex1.add_argument(
        "-b",
        "--begin-date",
        type=vdate,
        metavar="<YYYY-MM-DDTHH:MM:SS>",
        default=None,
        help="by begin date",
    )
    ex1.add_argument("-l", "--latest", action="store_true", help="latest closed batch")
    ex1.add_argument("-a", "--all", action="store_true", help="all closed batches")
    parser.add_argument("-d", "--base-dir", type=vdir, default=".", help="Base dir for the export")
    parser.add_argument("-e", "--email", action="store_true", help="Send results by email")
    parser.add_argument(
        "-u",
        "--updated",
        action="store_true",
        help="Do action only when ZP updated flag is True|False",
    )
    return parser


def trange() -> ArgumentParser:
    parser = ArgumentParser(add_help=False)
    parser.add_argument(
        "-s",
        "--since",
        type=vdate,
        metavar="<YYYY-MM-DDTHH:MM:SS>",
        default=None,
        help="since date, defaults to beginning of current year",
    )
    parser.add_argument(
        "-u",
        "--until",
        type=vdate,
        metavar="<YYYY-MM-DDTHH:MM:SS>",
        default=None,
        help="until date, defaults to end of current year",
    )
    return parser


def detailed() -> ArgumentParser:
    parser = ArgumentParser(add_help=False)
    parser.add_argument(
        "-d",
        "--detailed",
        action="store_true",
        help="Export a CSV with a summary (only) for a given time span",
    )
    return parser


def mode() -> ArgumentParser:
    parser = ArgumentParser(add_help=False)
    parser.add_argument(
        "-m",
        "--mode",
        type=Calibration,
        default=None,
        choices=Calibration,
        help="Calibration mode, defaults to %(default)s",
    )
    return parser


def ploto() -> ArgumentParser:
    parser = ArgumentParser(add_help=False)
    ex1 = parser.add_mutually_exclusive_group(required=False)
    ex1.add_argument("-sa", "--plot-samples", action="store_true", help="Ref/test samples vs time")
    ex1.add_argument("-hi", "--plot-histo", action="store_true", help="Ref/Test samples histogram")
    ex1.add_argument("-bo", "--plot-both", action="store_true", help="Ref/Test hist + samples vs time")
    parser.add_argument(
        "-me",
        "--median",
        action="store_true",
        help="Stats use median instead of mean",
    )
    return parser



from enum import StrEnum

from typing import Tuple, Sequence, Mapping

from lica.asyncio.photometer import Role


# Per-Role Round Statistics type. 
# Tuple[0] =frequency, 
# Tuple[1] = freq std dev, 
# Tuple[2] magnitude according to ficticios ZP
RoundStatistics = Tuple[float, float, float]

# Summary statictis
# Tuple[0]: sequence of zero points, 
# Tuple[1]: sequence of reference central frequencies
# Tuple[2]: sequence of test central frecuencies
SummaryStatistics = Tuple[Sequence[float], Sequence[float], Sequence[float]]

RoundStatsType = Mapping[Role, RoundStatistics]


class Event(StrEnum):
	READING = "reading_event"
	ROUND = "round_event"
	SUMMARY = "summary_event"
	CAL_START = "calib_start_event"
	CAL_END = "calib_end_event"




# ----------------------------------------------------------------------
# Copyright (c) 2024 Rafael Gonzalez.
#
# See the LICENSE file for details
# ----------------------------------------------------------------------

# --------------------
# System wide imports
# -------------------

import sys
import math
import logging
import asyncio
import contextlib
import statistics

from logging import Logger
from typing import Any

# -------------------
# Third party imports
# -------------------

from lica.asyncio.photometer import Role

# --------------
# local imports
# -------------

from ...controller.photometer import Controller


def mag(zp: float, freq_offset: float, freq: float):
    """Single sample magnitude calculation that takes into account zero frequencies"""
    f = freq - freq_offset
    return (float(zp) - 2.5 * math.log10(f)) if f > 0.0 else math.inf


def log_msgs_stats(msgs: list[dict[str, Any], ...], controller: Controller, role: Role, name: str) -> None:
    freqs = [msg["freq"] for msg in msgs]
    mean = statistics.fmean(freqs)
    median = statistics.median_low(freqs)
    stddev_mean = statistics.stdev(freqs, xbar=mean)
    stddev_median = statistics.stdev(freqs, xbar=median)
    multimode = statistics.multimode(freqs)
    zp = controller.phot_info[role]["zp"]
    fo = controller.phot_info[role]["freq_offset"]
    log = logging.getLogger(role.tag())
    log.info(
        "%-9s stats => mean freq = %.03f Hz, median = %.03f Hz, \u03c3(mean) = %.03f, \u03c3(median) = %.03f, local maxima = %s",
        name,
        mean,
        median,
        stddev_mean,
        stddev_median,
        multimode,
    )
    log.info(
        "%-9s stats => mean mag = %.02f @ %.02f",
        name,
        mag(zp, fo, mean),
        zp,
    )



async def log_phot_info(controller: Controller, role: Role) -> None:
    log = logging.getLogger(role.tag())
    phot_info = await controller.info(role)
    log.info("-" * 40)
    for key, value in sorted(phot_info.items()):
        log.info("%-12s: %s", key.upper(), value)
    log.info("-" * 40)


async def log_messages(controller: Controller, role: Role, num: int | None = None) -> list[dict[str,Any]]:
    log = logging.getLogger(role.tag())
    name = controller.phot_info[role]["name"]
    zp = controller.phot_info[role]["zp"]
    fo = controller.phot_info[role]["freq_offset"]
    # Although in this case, it doesn't matter, in general
    # async generatores may not close as expected,
    # hence the use of closing() context manager
    messages = list()
    async with contextlib.aclosing(controller.readings(role, num)) as generator:
        async for role, msg in generator:
            messages.append(msg)
            log.info(
                "%-9s [%d] T=%s, f=%s Hz, mag=%0.2f @ %s, tbox=%s, tsky=%s",
                name,
                msg.get("seq"),
                msg.get("tstamp"),
                msg["freq"],
                msg.get("mag", mag(zp, fo, msg["freq"])),
                zp,
                msg["tamb"],
                msg["tsky"],
            )
    log_msgs_stats(messages, controller, role, name)
    return messages



async def update_zp(controller: Controller, zero_point: float) -> None:
    log = logging.getLogger(Role.TEST.tag())
    zero_point = round(zero_point, 2)
    log.info("Updating ZP : %0.2f", zero_point)
    name = controller.phot_info[Role.TEST]["name"]
    try:
        stored_zp = await controller.write_zp(zero_point)
    except asyncio.exceptions.TimeoutError:
        log.error("[%s] Failed contacting %s photometer", name, Role.TEST.tag())
    except Exception as e:
        log.exception(e)
    else:
        if stored_zp is None:
            log.error(
                "ZP Write verification failed: ZP to Write (%0.3f) "
                "doesn't match ZP subsequently read (%s). \u0394 = ?",
                zero_point,
                stored_zp,
            )
        elif stored_zp is None or zero_point != stored_zp:
            log.error(
                "ZP Write verification failed: ZP to Write (%0.3f) "
                "doesn't match ZP subsequently read (%0.3f). \u0394 = %0.16f",
                zero_point,
                stored_zp,
                zero_point - stored_zp,
            )
        else:
            log.info("[%s] ZP Write (%0.2f) verification Ok.", name, zero_point)

def log_and_exit(log: Logger, excp: Exception, trace: bool) -> None:
    if trace:
        log.exception(excp)
    else:
        log.error(excp)
    sys.exit(1)


# ----------------------------------------------------------------------
# Copyright (c) 2024 Rafael Gonzalez.
#
# See the LICENSE file for details
# ----------------------------------------------------------------------

# --------------------
# System wide imports
# -------------------

from datetime import datetime
from math import log10
import statistics
from collections import Counter

# ---------------------------
# Third-party library imports
# ----------------------------

import decouple
import matplotlib
import matplotlib.pyplot as plt
from matplotlib.ticker import FormatStrFormatter
from lica.asyncio.photometer import Role

# --------------
# local imports
# -------------

from ..constants import ZP_ABS, FreqSequence, TimeSequence

# Load-time configuration
TK_AGG = decouple.config("MPL_USE_TK", default=False, cast=bool)
if TK_AGG:
    matplotlib.use("TkAgg")  # For very old laptops


def stats(
    series: list[float, ...], use_median: bool = False
) -> tuple[float, float, list[float, ...]]:
    """Compute mean(/median and std dev around central tendency"""
    central = statistics.median_low(series) if use_median else statistics.fmean(series)
    std_dev = statistics.stdev(series, xbar=central)
    modes = statistics.multimode(series)
    return central, std_dev, modes


def samples(
    session: datetime,
    roles: list[Role, ...],
    freqs: list[FreqSequence, ...],
    tstamps: list[TimeSequence, ...],
    names: list[str, ...],
    use_median: bool = False,
    zp_abs: float = ZP_ABS,
) -> None:
    """Grafica Frecuencia vs Tiempo en N rondas"""
    session_id = session.strftime("%Y-%m-%dT%H:%M:%S")
    fig, axes = plt.subplots(1, figsize=(12, 5))
    central = "median" if use_median else "mean"
    n = len(roles)
    axes = [
        axes,
    ] * n
    colors = ("tab:red", "tab:blue") if n == 2 else ("tab:red",)
    roles_dict = dict()
    for i, ax, role, freq, tstamp, name, color, role in zip(
        range(1, n + 1), axes, roles, freqs, tstamps, names, colors, roles
    ):
        cen, std, modes = stats(freq, use_median=use_median)
        roles_dict[role] = (cen, std, modes)
        ax.set_title(f"Session {session_id}")
        ax.set_ylabel("Frecuency (Hz)")
        # Medidas
        ax.plot(tstamp, freq, color=color, marker=".", linestyle="none", label=f"{name}")
        # Tendencias centrales
        ax.axhline(y=cen, linestyle=":", color=color, label=f"{central} = {cen:.3f}")
        # Barra horizontal semitransparente para la cota de error estimada (2 sigma)
        ax.axhspan(
            cen - 2 * std,
            cen + 2 * std,
            alpha=0.1,
            color=color,
            label=rf"$2\sigma$ ref = {std:.3f}",
        )
        ax.legend()
    if n == 2:
        # caja central de calibración de ronda
        delta_mag = -2.5 * log10(roles_dict[Role.REF][0] / roles_dict[Role.TEST][0])
        zp = zp_abs + delta_mag
        texto = rf"$ \Delta m = {delta_mag:.02f}$;  $ZP_{i} = \Delta m + ZP_{{abs}} = {zp:.02f}$"
        axes[0].annotate(
            texto,
            xy=(0.5, 0.5),
            xycoords="axes fraction",
            ha="center",
            va="center",
            bbox=dict(boxstyle="round", facecolor="lightblue", alpha=0.3),
        )
    plt.tight_layout()
    plt.show()


def histograms(
    session: datetime,
    roles: list[Role, ...],
    freqs: list[FreqSequence, ...],
    tstamps: list[FreqSequence, ...],
    names: list[str, ...],
    use_median: bool = False,
    title: str | None = None,
    subtitles: list[str, ...] | None = None,
    labels: list[str, ...] | None = None,
    decimals: list[int, ...] = None,
) -> None:
    """
    Histogram for the ref and test photometer frequencies.
    One row, two couluns
    """
    # Crear figura con 1 fila, n columnas para uno/dos histogramas
    n = len(roles)
    fig, axes = plt.subplots(1, n, figsize=(12, 5))
    if n == 2:
        title = title or f"Histograms of {names[0]} & {names[1]} on {session}"
    else:
        title = title or f"Histograms of {names[0]} on {session}"
        axes = [axes]
    fig.suptitle(title)
    central = "median" if use_median else "mean"
    subtitles = subtitles or [None] * n
    labels = labels or [None] * n
    decimals = decimals or [3] * n
    distributions = list()
    allstats = list()
    for i in range(n):
        distributions.append(Counter([round(f, decimals[i]) for f in freqs[i]]))
        allstats.append(stats(freqs[i], use_median=use_median))

    for i, (ax, distr, name, label, subtitle, mystats, decim) in enumerate(
        zip(axes, distributions, names, labels, subtitles, allstats, decimals)
    ):
        x = list(distr.keys())  # Clave del Counter
        y = list(distr.values())  # Cuenta del Counter
        total = sum(y)
        width = 10 ** (-decim)
        cen, stddev, modes = mystats[0], mystats[1], mystats[2]
        ax.axvline(cen, color="red", linestyle="--", label=f"{central} = {cen:.3f}")
        for i, mode in enumerate(modes, start=1):
            ax.axvline(mode, color="red", linestyle=":", label=f"mode {i}= {mode:.3f}")
        ax.axvspan(
            cen - 2 * stddev,
            cen + 2 * stddev,
            alpha=0.1,
            color="blue",
            label=rf"$2\sigma$ = {stddev:.3f}",
        )
        ax.bar(x, y, width=width, alpha=0.8)
        ax.set_title(name)
        if label is not None:
            ax.set_xlabel(label)
        ax.set_ylabel(f"Counts ({total} Total)")
        ax.legend()
        ax.xaxis.set_major_formatter(FormatStrFormatter(f"% .{decim}f"))
        ax.set_xticks(x)
        ax.grid(True, alpha=0.3)
    plt.tight_layout()
    plt.show()




# ---------------------
# Third party libraries
# ---------------------

from sqlalchemy import select
from zptessdao.asyncio import Batch

# -------------
# Own Libraries
# -------------

from ...dao import Session

# Re-exports
from .batch import Controller as Controller

async def get_open_batch(session: Session) -> Batch | None:
        """Used by the persistent controller"""
        q = select(Batch).where(Batch.end_tstamp.is_(None))
        batch = (await session.scalars(q)).one_or_none()
        return batch


# ----------------------------------------------------------------------
# Copyright (c) 2024 Rafael Gonzalez.
#
# See the LICENSE file for details
# ----------------------------------------------------------------------

# --------------------
# System wide imports
# -------------------
from __future__ import annotations

import logging
import asyncio
from collections import defaultdict

from typing import Any, Mapping, Dict, List


# ---------------------------
# Third-party library imports
# ----------------------------

from sqlalchemy import select
from pubsub import pub
from lica.asyncio.photometer import Role
from zptessdao.asyncio import Photometer, Summary, Round, Sample
from zptessdao.constants import Calibration

# --------------
# local imports
# -------------

from ...dao import Session
from ..batch import get_open_batch
from .volatile import Controller as VolatileCalibrator
from .types import Event

from ... import __version__

# ----------------
# Module constants
# ----------------

# -----------------------
# Module global variables
# -----------------------

# get the module logger
log = logging.getLogger(__name__.split(".")[-1])

# -------------------
# Auxiliary functions
# -------------------

# -----------------
# Auxiliary classes
# -----------------


class Controller(VolatileCalibrator):
    """
    Database-based Photometer Calibration Controller
    """

    def __init__(
        self,
        ref_params: Mapping[str, Any] | None = None,
        test_params: Mapping[str, Any] | None = None,
        common_params: Mapping[str, Any] | None = None,
    ):
        super().__init__(ref_params, test_params, common_params)
        self.db_queue = asyncio.Queue()
        self.batch = None

    # ==========
    # Public API
    # ==========

    async def init(self) -> None:
        await super().init()
        async with Session() as session:
            self.batch = await get_open_batch(session)
        self.db_task = asyncio.create_task(self.db_writer_task())

    async def calibrate(self) -> float:
        zp = await super().calibrate()
        await asyncio.wait([self.db_task])
        return zp

    async def write_zp(self, zero_point: float) -> float:
        """May raise asyncio.exceptions.TimeoutError in particular"""
        stored_zero_point = None
        try:
            stored_zero_point = await super().write_zp(zero_point)
        except Exception as e:
            log.error(e)
            updated = False
        else:
            updated = True
        async with Session() as session:
            async with session.begin():
                q = select(Summary).where(Summary.session == self.meas_session)
                db_summaries = (await session.scalars(q)).all()
                for db_summary in db_summaries:
                    db_summary.upd_flag = False if db_summary.role == Role.REF else updated
                    if not updated:
                        db_summary.comment = f"{self.phot_info[Role.TEST]['name']} not updated because of HTTP Timeout"
        return stored_zero_point

    async def not_updated(self, zero_point: float, msg: str):
        """What to do when the Zero Point is not updated by the client code"""
        async with Session() as session:
            async with session.begin():
                q = select(Summary).where(Summary.session == self.meas_session)
                db_summaries = (await session.scalars(q)).all()
                for db_summary in db_summaries:
                    db_summary.upd_flag = False
                    db_summary.comment = msg

    # ===========
    # Private API
    # ===========

    # --------------------
    # Hooks implementation
    # --------------------

    def _on_calib_start(self) -> None:
        pub.sendMessage(Event.CAL_START)
        msg = {"event": Event.CAL_START, "info": None}
        self.db_queue.put_nowait(msg)

    def _on_calib_end(self) -> None:
        pub.sendMessage(Event.CAL_END)
        msg = {"event": Event.CAL_END, "info": None}
        self.db_queue.put_nowait(msg)

    def _on_round(self, round_info: Mapping[str, Any]) -> None:
        pub.sendMessage(Event.ROUND, **round_info)
        # We must copy the sequence of samples of a given round
        # since the background filling tasks are active
        msg = {
            "event": Event.ROUND,
            "info": round_info,
        }
        self.db_queue.put_nowait(msg)

    def _on_summary(self, summary_info: Mapping[str, Any]) -> None:
        pub.sendMessage(Event.SUMMARY, **summary_info)
        msg = {"event": Event.SUMMARY, "info": summary_info}
        self.db_queue.put_nowait(msg)

    # ----------------------------------
    # Coroutines to be turned into Tasks
    # ----------------------------------

    async def db_writer_task(self) -> None:
        self.db_active = True
        self.temp_round_info = list()
        self.temp_round_samples = list()
        while self.db_active:
            msg = await self.db_queue.get()
            event = msg["event"]
            if event == Event.CAL_START:
                pass
            elif event == Event.ROUND:
                self.temp_round_info.append(msg["info"])
            elif event == Event.SUMMARY:
                self.temp_summary = msg["info"]
            else:
                try:
                    await self._save_all()
                except Exception as e:
                    log.error(e)
                    log.critical(
                        "Probably an incomplete manual purge forgot table samples_rounds_t references"
                    )

    # ----------------------
    # Private helper methods
    # ----------------------

    async def _save_photometers(self, session: Session) -> Dict[Role, Photometer]:
        phot = dict()
        for role in self.roles:
            name = self.phot_info[role]["name"]
            mac = self.phot_info[role]["mac"]
            q = select(Photometer).where(Photometer.mac == mac, Photometer.name == name)
            phot[role] = (await session.scalars(q)).one_or_none()
            if phot[role] is None:
                col = dict()
                for key in ("name", "mac", "model", "sensor", "freq_offset", "firmware"):
                    col[key] = self.phot_info[role][key] or None
                col["freq_offset"] = col["freq_offset"] or 0.0
                phot[role] = Photometer(**col)
                session.add(phot[role])
        return phot

    def _save_summaries(
        self, session: Session, photometers: Dict[Role, Photometer]
    ) -> Dict[Role, Summary]:
        db_summary = dict()
        for role, phot in photometers.items():
            db_summary[role] = Summary(
                session=self.meas_session,
                role=role,
                calibration=Calibration.AUTO,
                calversion=__version__,
                author=self.author,
                zp_offset=self.zp_offset if role == Role.TEST else 0,
                prev_zp=self.phot_info[role]["zp"] if role == Role.TEST else self.zp_abs,
                zero_point=self.temp_summary["best_zero_point"]
                if role == Role.TEST
                else self.zp_abs,
                zero_point_method=self.temp_summary["best_zero_point_method"]
                if role == Role.TEST
                else None,
                freq=self.temp_summary["best_freq"][role],
                freq_method=self.temp_summary["best_freq_method"][role],
                mag=self.temp_summary["best_mag"][role],
                nrounds=self.nrounds,
                photometer=phot,  # This is really a many to one relationship
                batch=self.batch,  # Optional many-to-one relationships (NULLS are allowed)
            )
            session.add(db_summary[role])
        return db_summary

    def _save_rounds(
        self, session: Session, db_summaries: Dict[Role, Summary]
    ) -> Dict[Role, List[Round]]:
        db_rounds = defaultdict(list)
        for i, round_info in enumerate(self.temp_round_info):
            for role, summary in db_summaries.items():
                samples = self.accum_samples[role][i]
                tstamps = self.time_intervals[role][i]
                r = Round(
                    seq=round_info["current"],
                    role=role,
                    freq=round_info["stats"][role][0],
                    stddev=round_info["stats"][role][1],
                    mag=round_info["stats"][role][2],
                    central=self.central,
                    zp_fict=self.zp_fict,
                    zero_point=round_info["zero_point"] if role == Role.TEST else None,
                    nsamples=len(samples),
                    begin_tstamp=tstamps[0],
                    end_tstamp=tstamps[1],
                    duration=(tstamps[1] - tstamps[0]).total_seconds(),
                    summary=summary,  # This is really a 1:N relationship
                )
                db_rounds[role].append(r)
                log.debug(r)
                session.add(r)
        return db_rounds

    def _save_samples(
        self,
        session: Session,
        db_summaries: Dict[Role, Summary],
        db_rounds: Dict[Role, List[Round]],
    ) -> Dict[Role, List[Sample]]:
        db_samples = dict()
        samples = defaultdict(set)
        # This double loop accumulate unique samples
        # dispersed in the rounds
        for role, summary in db_summaries.items():
            for q in self.accum_samples[role]:
                samples[role].update(q)
            db_samples[role] = [
                Sample(
                    tstamp=sample["tstamp"],
                    role=role,
                    seq=sample["seq"],
                    freq=sample["freq"],
                    temp_box=sample["tamb"],
                    summary=summary,
                )
                for sample in samples[role]
            ]
            # Adding the database sample objects to the summary
            for s in db_samples[role]:
                log.debug(s)
                session.add(s)
        # Now assign the samples to the corresponding round
        # The final list of unique samples is tested against the list of round samples
        # and added to the database round object if so.
        # A bit tricky (3-level for loop)
        for role in self.roles:
            for sample, db_sample in zip(samples[role], db_samples[role]):
                for i, db_round in enumerate(db_rounds[role]):
                    if sample in self.accum_samples[role][i]:
                        db_round.samples.append(db_sample)
        return db_samples

    async def _save_all(self):
        async with Session() as session:
            async with session.begin():
                db_photometers = await self._save_photometers(session)
                log.info("Saving %d photometer entries", len(db_photometers))
                log.debug(db_photometers)
                db_summaries = self._save_summaries(session, db_photometers)
                log.info("Saving %d summary entries", len(db_summaries))
                log.debug(db_summaries)
                db_rounds = self._save_rounds(session, db_summaries)
                log.info("Saving %d %s round entries", len(db_rounds[Role.REF]), Role.REF)
                log.info("Saving %d %s round entries", len(db_rounds[Role.TEST]), Role.TEST)
                db_samples = self._save_samples(session, db_summaries, db_rounds)
                log.info("Saving %d %s sample entries", len(db_samples[Role.REF]), Role.REF)
                log.info("Saving %d %s sample entries", len(db_samples[Role.TEST]), Role.TEST)
        self.db_active = False



# ----------------------------------------------------------------------
# Copyright (c) 2024 Rafael Gonzalez.
#
# See the LICENSE file for details
# ----------------------------------------------------------------------

# --------------------
# System wide imports
# -------------------

import logging
import asyncio
import statistics
from datetime import timedelta
from argparse import Namespace, ArgumentParser
from typing import Sequence, Mapping

# -------------------
# Third party imports
# -------------------

from pubsub import pub

from lica.sqlalchemy import sqa_logging
from lica.asyncio.cli import execute
from lica.asyncio.photometer import Role, Message

from zptessdao.constants import CentralTendency


# --------------
# local imports
# -------------

from .. import __version__
from .util import parser as prs
from .util.misc import log_phot_info, update_zp
from ..controller.photometer import VolatileCalibrator, PersistentCalibrator, Event, RoundStatsType
from ..controller.batch import Controller as BatchController
from ..dao import engine
from ..mpl import plot

# ----------------
# Module constants
# ----------------


DESCRIPTION = "TESS-W Reader tool"
HALF_SECOND = timedelta(seconds=0.5)

# -----------------------
# Module global variables
# -----------------------

# get the module logger
log = logging.getLogger(__name__.split(".")[-1])

# ------------------
# Auxiliar functions
# ------------------


def on_reading(role: Role, reading: Message) -> None:
    global controller
    log = logging.getLogger(role.tag())
    current = len(controller.buffer(role))
    total = controller.buffer(role).capacity()
    name = controller.phot_info[role]["name"]
    if current < total:
        log.info("%-9s waiting for enough samples, %03d remaining", name, total - current)


def on_round(current: int, mag_diff: float, zero_point: float, stats: RoundStatsType) -> None:
    global controller
    zp_abs = controller.zp_abs
    nrounds = controller.nrounds
    phot_info = controller.phot_info
    central = controller.central
    zp_fict = controller.zp_fict
    log.info("=" * 74)
    log.info(
        "%-10s %02d/%02d: New ZP = %0.2f = \u0394(ref-test) Mag (%0.2f) + ZP Abs (%0.2f)",
        "ROUND",
        current,
        nrounds,
        zero_point,
        mag_diff,
        zp_abs,
    )
    for role in (Role.REF, Role.TEST):
        tag = role.tag()
        name = phot_info[role]["name"]
        Ti = controller.ring[role][0]["tstamp"]
        Tf = controller.ring[role][-1]["tstamp"]
        T = (Tf - Ti).total_seconds()
        Ti = (Ti + HALF_SECOND).strftime("%H:%M:%S")
        Tf = (Tf + HALF_SECOND).strftime("%H:%M:%S")
        N = len(controller.ring[role])
        freq, stdev, mag = stats[role]
        log.info(
            "[%s] %-8s (%s-%s)[%4.1fs][%03d] %6s f = %0.3f Hz, \u03c3 = %0.3f Hz, m = %0.2f @ %0.2f",
            tag,
            name,
            Ti,
            Tf,
            T,
            N,
            central,
            freq,
            stdev,
            mag,
            zp_fict,
        )
    if current == nrounds:
        log.info("=" * 74)


def on_summary(
    zero_point_seq: Sequence[float],
    freq_seq: Mapping[Role, Sequence[float]],
    best_freq: Mapping[Role, float],
    best_freq_method: Mapping[Role, CentralTendency],
    best_mag: Mapping[Role, float],
    mag_diff: float,
    best_zero_point: float,
    best_zero_point_method: CentralTendency,
    final_zero_point: float,
    overlapping_windows: Mapping[Role, Sequence[float | None]],
) -> None:
    global controller
    log.info("#" * 74)
    log.info("Session = %s", controller.meas_session.strftime("%Y-%m-%dT%H:%M:%S"))
    log.info("Best ZP        list is %s", zero_point_seq)
    log.info("Best REF. Freq list is %s", freq_seq[Role.REF])
    log.info("Best TEST Freq list is %s", freq_seq[Role.TEST])
    log.info(
        "REF. Best Freq. = %0.3f Hz, Mag. = %0.2f, Diff %0.2f (%s)",
        best_freq[Role.REF],
        best_mag[Role.REF],
        0,
        best_freq_method[Role.REF],
    )
    log.info(
        "TEST Best Freq. = %0.3f Hz, Mag. = %0.2f, Diff %0.2f (%s)",
        best_freq[Role.TEST],
        best_mag[Role.TEST],
        mag_diff,
        best_freq_method[Role.TEST],
    )
    log.info(
        "Final TEST ZP (%0.2f) = Best ZP (%0.2f) (%s) + ZP offset (%0.2f)",
        final_zero_point,
        best_zero_point,
        best_zero_point_method,
        controller.zp_offset,
    )
    log.info(
        "Old TEST ZP = %0.2f, NEW TEST ZP = %0.2f",
        controller.phot_info[Role.TEST]["zp"],
        final_zero_point,
    )
    log.info("REF. rounds overlap \u0394T = %s", overlapping_windows[Role.REF])
    log.info("TEST rounds overlap \u0394T = %s", overlapping_windows[Role.TEST])
    log.info("REF. unique samples: %d", len(controller.unique_samples(Role.REF)))
    log.info("TEST unique samples: %d", len(controller.unique_samples(Role.TEST)))
    log.info("#" * 74)


# -----------------
# Auxiliary classes
# -----------------


# -------------------
# Auxiliary functions
# -------------------


async def cli_calib_test(args: Namespace) -> None:
    global controller

    ref_params = {
        "model": args.ref_model,
        "sensor": args.ref_sensor,
        "endpoint": args.ref_endpoint,
        "old_proto": args.ref_old_proto,
        "log_level": logging.DEBUG if args.ref_raw_message else logging.INFO,
        "strict": args.ref_strict,
    }
    test_params = {
        "model": args.test_model,
        "sensor": args.test_sensor,
        "endpoint": args.test_endpoint,
        "old_proto": args.test_old_proto,
        "log_level": logging.DEBUG if args.test_raw_message else logging.INFO,
        "strict": args.test_strict,
    }
    common_params = {
        "buffer": args.buffer,
        "persist": args.persist,
        "update": args.update,
        "central": args.central,
        "period": args.period,
        "zp_fict": args.zp_fict,
        "zp_offset": args.zp_offset,
        "rounds": args.rounds,
        "author": " ".join(args.author) if args.author else None,
    }
    if args.persist:
        controller = PersistentCalibrator(
            ref_params=ref_params, test_params=test_params, common_params=common_params
        )
        open_batch = await (BatchController()).is_open()
        if not open_batch:
            if args.no_batch:
                log.warn("Persistent calibration without an open batch")
            else:
                raise RuntimeError("Persistent calibration without an open batch")
        else:
            batch = await (BatchController()).get_open()
            log.info("Logging results to a database. Current batch is %s", batch)

    else:
        controller = VolatileCalibrator(
            ref_params=ref_params, test_params=test_params, common_params=common_params
        )
    pub.subscribe(on_reading, Event.READING)
    pub.subscribe(on_round, Event.ROUND)
    pub.subscribe(on_summary, Event.SUMMARY)

    await controller.init()
    try:
        async with asyncio.TaskGroup() as tg:
            tg.create_task(log_phot_info(controller, Role.REF))
            tg.create_task(log_phot_info(controller, Role.TEST))
    except* Exception as eg:
        for e in eg.exceptions:
            if args.trace:
                log.exception(e)
            else:
                log.error(e)
        raise RuntimeError("Could't continue execution, check errors above")
    if args.info:
        log.info("Only displaying info. Stopping here.")
        return
    final_zero_point = await controller.calibrate()
    if args.update:
        await update_zp(controller, final_zero_point)
    else:
        msg = f"Zero Point {final_zero_point:.2f} not saved to {Role.TEST} {controller.phot_info[Role.TEST]['name']}"
        log.info(msg)
        await controller.not_updated(final_zero_point, msg)
    ref_freqs = [msg["freq"] for msg in controller.unique_samples(Role.REF)]
    ref_tstamps = [msg["tstamp"] for msg in controller.unique_samples(Role.REF)]
    tst_freqs = [msg["freq"] for msg in controller.unique_samples(Role.TEST)]
    tst_tstamps = [msg["tstamp"] for msg in controller.unique_samples(Role.TEST)]
    ref_name = controller.phot_info[Role.REF]["name"]
    tst_name = controller.phot_info[Role.TEST]["name"]
    decimals = 2 if statistics.mean(ref_freqs) > 3 else 3
    if args.plot_histo:
        plot.histograms(
            session=controller.meas_session,
            roles=[Role.REF, Role.TEST],
            freqs=[ref_freqs, tst_freqs],
            tstamps=[ref_tstamps, tst_tstamps],
            names=[ref_name, tst_name],
            decimals=[decimals,2]
        )
    elif args.plot_samples:
        plot.samples(
            session=controller.meas_session,
            roles=[Role.REF, Role.TEST],
            freqs=[ref_freqs, tst_freqs],
            tstamps=[ref_tstamps, tst_tstamps],
            names=[ref_name, tst_name],
        )
    elif args.plot_both:
        plot.histograms(
            session=controller.meas_session,
            roles=[Role.REF, Role.TEST],
            freqs=[ref_freqs, tst_freqs],
            tstamps=[ref_tstamps, tst_tstamps],
            names=[ref_name, tst_name],
            decimals=[decimals,2]
        )
        plot.samples(
            session=controller.meas_session,
            roles=[Role.REF, Role.TEST],
            freqs=[ref_freqs, tst_freqs],
            tstamps=[ref_tstamps, tst_tstamps],
            names=[ref_name, tst_name],
        )





# -----------------
# CLI API functions
# -----------------


def add_args(parser: ArgumentParser):
    subparser = parser.add_subparsers(dest="command", required=True)
    p = subparser.add_parser(
        "test",
        parents=[
            prs.info(),
            prs.stats(),
            prs.upd(),
            prs.persist(),
            prs.buf(),
            prs.author(),
            prs.ref(),
            prs.test(),
            prs.no_bat(),
            prs.ploto(),
        ],
        help="Calibrate test photometer",
    )
    p.set_defaults(func=cli_calib_test)


async def cli_main(args: Namespace) -> None:
    sqa_logging(args)
    await args.func(args)
    await engine.dispose()


def main():
    """The main entry point specified by pyproject.toml"""
    execute(
        main_func=cli_main,
        add_args_func=add_args,
        name=__name__,
        version=__version__,
        description=DESCRIPTION,
    )


if __name__ == "__main__":
    main()



# ----------------------------------------------------------------------
# Copyright (c) 2024 Rafael Gonzalez.
#
# See the LICENSE file for details
# ----------------------------------------------------------------------

# --------------------
# System wide imports
# -------------------

import sys
from abc import ABC, abstractmethod
import logging
import asyncio
from typing import Mapping, Any, Dict, Tuple, AsyncIterator


# ---------------------------
# Third-party library imports
# ----------------------------

from lica.asyncio.photometer import Role, Message as PhotMessage, Model as PhotModel, Sensor

# --------------
# local imports
# -------------

from .. import load_config
from ...dao import engine, Session
from .builder import PhotometerBuilder

# ----------------
# Module constants
# ----------------


SECTION = {Role.REF: "ref-device", Role.TEST: "test-device"}

# -----------------------
# Module global variables
# -----------------------

# get the module logger
log = logging.getLogger(__name__.split(".")[-1])

# -------------------
# Auxiliary functions
# -------------------

# -----------------
# Auxiliary classes
# -----------------


class Controller(ABC):
    """
    Reader Controller specialized in reading the photometers
    Serves ans an interface and a base class at the same time.
    """

    def __init__(
        self,
        ref_params: Mapping[str, Any] | None = None,
        test_params: Mapping[str, Any] | None = None,
    ):
        self.param = {Role.REF: ref_params, Role.TEST: test_params}
        self.roles = list()
        self.photometer = dict()
        self.ring = dict()
        self.phot_info = dict()
        self.phot_task = dict()
        if ref_params is not None:
            self.roles.append(Role.REF)
        if test_params is not None:
            self.roles.append(Role.TEST)

    # ==========
    # Public API
    # ==========

    def buffer(self, role: Role):
        return self.ring[role]

    async def init(self) -> None:
        log.info(
            "Initializing %s controller for %s",
            self.__class__.__name__,
            self.roles,
        )
        # Use engine parameter for the reference photometer when using database info
        builder = PhotometerBuilder(engine)
        async with Session() as session:
            for role in self.roles:
                val_db = await load_config(session, SECTION[role], "model")
                val_arg = self.param[role]["model"]
                self.param[role]["model"] = val_arg if val_arg is not None else PhotModel(val_db)
                val_db = await load_config(session, SECTION[role], "sensor")
                val_arg = self.param[role]["sensor"]
                self.param[role]["sensor"] = val_arg if val_arg is not None else Sensor(val_db)
                val_db = await load_config(session, SECTION[role], "old-proto")
                val_arg = self.param[role]["old_proto"]
                self.param[role]["old_proto"] = val_arg if val_arg is not None else bool(val_db)
                val_db = await load_config(session, SECTION[role], "endpoint")
                val_arg = self.param[role]["endpoint"]
                self.param[role]["endpoint"] = val_arg if val_arg is not None else val_db
                self.photometer[role] = builder.build(
                    self.param[role]["model"],
                    role,
                    self.param[role]["endpoint"],
                    self.param[role]["strict"],
                )
                logging.getLogger(str(role)).setLevel(self.param[role]["log_level"])

    async def info(self, role: Role) -> Dict[str, Any]:
        log = logging.getLogger(role.tag())
        try:
            phot_info = await self.photometer[role].get_info()
        except asyncio.exceptions.TimeoutError:
            log.error("Failed contacting %s photometer", role.tag())
            raise
        else:
            phot_info["endpoint"] = role.endpoint()
            phot_info["sensor"] = phot_info["sensor"] or self.param[role]["sensor"].value
            v = phot_info["freq_offset"] or 0.0
            phot_info["freq_offset"] = float(v)
            phot_info["zp"] = float(phot_info["zp"]) # ref photometer zp has a string, coming from database
            self.phot_info[role] = phot_info
            return phot_info

    async def readings(
        self, role: Role, num_messages: int | None = None
    ) -> AsyncIterator[Tuple[Role, PhotMessage]]:
        """An asynchronous generator, to be used by clients with async for"""
        num_messages = num_messages or sys.maxsize
        i = 0
        async with self.photometer[role]:
            while i < num_messages:
                reading = await anext(self.photometer[role].readings)
                if reading is not None:
                    i += 1
                    yield role, reading

    async def write_zp(self, zero_point: float) -> float:
        """May raise asyncio.exceptions.TimeoutError in particular"""
        await self.photometer[Role.TEST].save_zero_point(zero_point)
        stored_zero_point = (await self.photometer[Role.TEST].get_info())["zp"]
        return stored_zero_point

    @abstractmethod
    async def calibrate(self) -> float:
        """Calibrate the test photometer against the refrence photometer returnoing a Zero Point"""
        pass



# ----------------------------------------------------------------------
# Copyright (c) 2024 Rafael Gonzalez.
#
# See the LICENSE file for details
# ----------------------------------------------------------------------

# --------------------
# System wide imports
# -------------------

import logging
from argparse import Namespace, ArgumentParser

# -------------------
# Third party imports
# -------------------

from lica.sqlalchemy import sqa_logging
from lica.asyncio.cli import execute
from lica.asyncio.photometer import Role

# --------------
# local imports
# -------------

from .. import __version__
from .util import parser as prs
from ..dao import engine
from ..controller.dbsamples import Controller as Sampler
from ..mpl import plot

# ----------------
# Module constants
# ----------------

TSTAMP_FMT = "%Y-%m-%dT%H:%M:%S"

# -----------------------
# Module global variables
# -----------------------

# get the module logger
log = logging.getLogger(__name__.split(".")[-1])


# -----------------
# Auxiliar function
# -----------------


# -----------------
# CLI API functions
# -----------------


async def cli_plot_session(args: Namespace) -> None:
    session = args.session
    sampler = Sampler()
    ref_freqs, ref_tstamps, ref_name = await sampler.samples(session, Role.REF)
    tst_freqs, tst_tstamps, tst_name = await sampler.samples(session, Role.TEST)
    ref_name, tst_name = ref_name[0], tst_name[0]
    if args.samples:
        plot.samples(
            session=session,
            roles=[Role.REF, Role.TEST],
            freqs=[ref_freqs, tst_freqs],
            tstamps=[ref_tstamps, tst_tstamps],
            names=[ref_name, tst_name],
            use_median=args.median,
        )
    elif args.histo:
        plot.histograms(
            session=session,
            roles=[Role.REF, Role.TEST],
            freqs=[ref_freqs, tst_freqs],
            tstamps=[ref_tstamps, tst_tstamps],
            names=[ref_name, tst_name],
            use_median=args.median,
        )
    elif args.both:
        plot.samples(
            session=session,
            roles=[Role.REF, Role.TEST],
            freqs=[ref_freqs, tst_freqs],
            tstamps=[ref_tstamps, tst_tstamps],
            names=[ref_name, tst_name],
            use_median=args.median,
        )

        plot.histograms(
            session=session,
            roles=[Role.REF, Role.TEST],
            freqs=[ref_freqs, tst_freqs],
            tstamps=[ref_tstamps, tst_tstamps],
            names=[ref_name, tst_name],
            use_median=args.median,
        )
    else:
        pass

    return


def add_args(parser: ArgumentParser):
    subparser = parser.add_subparsers(dest="command", required=True)
    p = subparser.add_parser(
        "session",
        parents=[prs.sess(), prs.ploto()],
        help="Plot calibration session samples",
    )
    p.set_defaults(func=cli_plot_session)


async def cli_main(args: Namespace) -> None:
    sqa_logging(args)
    await args.func(args)
    await engine.dispose()


def main():
    """The main entry point specified by pyproject.toml"""
    execute(
        main_func=cli_main,
        add_args_func=add_args,
        name=__name__,
        version=__version__,
        description="Plotting tools",
    )



# ----------------------------------------------------------------------
# Copyright (c) 2024 Rafael Gonzalez.
#
# See the LICENSE file for details
# ----------------------------------------------------------------------

# --------------------
# System wide imports
# -------------------

import logging

from typing import Any, Mapping

# ---------------------------
# Third-party library imports
# ----------------------------

from lica.asyncio.photometer import Role

# --------------
# local imports
# -------------

from .ring import RingBuffer
from .base import Controller as BaseController


# ----------------
# Module constants
# ----------------


SECTION = {Role.REF: "ref-device", Role.TEST: "test-device"}

# -----------------------
# Module global variables
# -----------------------

# get the module logger
log = logging.getLogger(__name__.split(".")[-1])

# -------------------
# Auxiliary functions
# -------------------

# -----------------
# Auxiliary classes
# -----------------


class Controller(BaseController):
    """
    Writer Controller specialized in writing Zero Points via API
    """

    def __init__(
        self,
        ref_params: Mapping[str, Any] | None = None,
        test_params: Mapping[str, Any] | None = None,
    ):
        super().__init__(ref_params, test_params)

    async def calibrate(self) -> float:
        """Calibrate the test photometer against the refrence photometer retirnoing a Zero Point"""
        raise NotImplementedError("Not relevant method for %s" % (self.__class__.__name__))

    async def init(self) -> None:
        await super().init()
        for role in self.roles:
            self.ring[role] = RingBuffer(capacity=1)
